# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for primitive_equations."""
from absl.testing import absltest
from absl.testing import parameterized

from dinosaur import coordinate_systems
from dinosaur import primitive_equations
from dinosaur import primitive_equations_states
from dinosaur import scales
from dinosaur import sigma_coordinates
from dinosaur import spherical_harmonic
from dinosaur import time_integration
from dinosaur import xarray_utils

import jax
from jax import config
import jax.numpy as jnp
import numpy as np


config.parse_flags_with_absl()

units = scales.units


def random_state(coords, key):
  (vorticity_key,
   divergence_key,
   temperature_variation_key,
   log_surface_pressure_key) = jax.random.split(key, 4)
  # All values are scaled by 1 / total_wavenumber**2
  scale = (coords.horizontal.total_wavenumbers + 1) ** -2
  vorticity = scale * jax.random.normal(
      vorticity_key,
      shape=coords.modal_shape)
  divergence = scale * jax.random.normal(
      divergence_key,
      shape=coords.modal_shape)
  temperature_variation = scale * jax.random.normal(
      temperature_variation_key,
      shape=coords.modal_shape)
  log_surface_pressure = scale * jax.random.normal(
      log_surface_pressure_key,
      shape=coords.surface_modal_shape)
  state = primitive_equations.State(vorticity,
                                    divergence,
                                    temperature_variation,
                                    log_surface_pressure)
  primitive_equations.validate_state_shape(state, coords)
  return state


def assert_states_close(state0, state1, **kwargs):
  for field in state0.fields:
    if field.name == 'tracers':
      for tracer_name in state0.tracers.keys():
        np.testing.assert_allclose(
            state0.tracers[tracer_name], state1.tracers[tracer_name],
            err_msg=f'Mismatch in tracer {tracer_name}:', **kwargs)
    else:
      if field.name == 'sim_time':
        if state0.sim_time is None != state1.sim_time is None:
          raise AssertionError(
              f'Mismatch is sim_time: {state0.sim_time} != {state1.sim_time}'
          )
        if state0.sim_time is None:  # assert_allclose does not handle None
          continue
      np.testing.assert_allclose(getattr(state0, field.name),
                                 getattr(state1, field.name),
                                 err_msg=f'Mismatch in {field}:',
                                 **kwargs)


class PrimitiveEquationsImplicitTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(
          wavenumbers=256,
          test_m_fn=lambda lon, lat: jnp.sin(lon) * jnp.cos(lat) ** 2,
          test_n_fn=lambda lon, lat: jnp.cos(lat) ** 2,
      ),
      dict(
          wavenumbers=128,
          test_m_fn=lambda lon, lat: 2.3 * jnp.cos(lon) ** 2 * jnp.cos(lat),
          test_n_fn=lambda lon, lat: 3.6 * jnp.cos(lat) * jnp.sin(2 * lat),
      ),
  )
  def testDivSecLat(self, wavenumbers, test_m_fn, test_n_fn):
    """Test that helper function div_sec_lat returns expected values."""
    grid = spherical_harmonic.Grid.with_wavenumbers(wavenumbers)
    lon, sin_lat = grid.nodal_mesh
    lat = np.arcsin(sin_lat)
    m = test_m_fn(lon, lat)
    n = test_n_fn(lon, lat)
    # should be same as H(M, N) = (1 / cos²θ) * ∂M/∂λ + (1 / cosθ) * ∂N/∂θ
    dm_dlon_fn = jax.vmap(jax.vmap(jax.grad(test_m_fn)))
    dn_dlat_fn = jax.vmap(jax.vmap(jax.grad(test_n_fn, argnums=1)))
    h_mn_expected = grid.to_modal(
        dm_dlon_fn(lon, lat) / (np.cos(lat) ** 2) +
        dn_dlat_fn(lon, lat) / np.cos(lat))
    h_mn_actual = primitive_equations.div_sec_lat(m, n, grid)
    np.testing.assert_allclose(h_mn_actual, h_mn_expected, atol=1e-3)

  @parameterized.parameters(
      dict(coordinates=sigma_coordinates.SigmaCoordinates.equidistant(10)),
      dict(coordinates=sigma_coordinates.SigmaCoordinates.equidistant(111)),
  )
  def testGetSigmaRatios(self, coordinates):
    """Tests that the values of the sigma ratios 𝛼 are correct."""
    alpha = primitive_equations.get_sigma_ratios(coordinates)
    np.testing.assert_array_equal([coordinates.layers], alpha.shape)
    sigma = coordinates.centers
    for j in range(coordinates.layers):
      if j == coordinates.layers - 1:
        expected_entry = -np.log(sigma[j])
      else:
        expected_entry = (np.log(sigma[j + 1]) - np.log(sigma[j])) / 2
      np.testing.assert_almost_equal(expected_entry, alpha[j])

  @parameterized.parameters(
      dict(wavenumbers=8, layers=3, atol=5e-3),
      dict(wavenumbers=128, layers=16, atol=5e-3),
  )
  def testGetGeopotentialSteadyState(self, wavenumbers, layers, atol):
    """Tests that `get_geopotential` works for steady states."""
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    coords = coordinate_systems.CoordinateSystem(
        horizontal=spherical_harmonic.Grid.with_wavenumbers(wavenumbers),
        vertical=sigma_coordinates.SigmaCoordinates.equidistant(layers))
    initial_state_fn, aux_features = primitive_equations_states.steady_state_jw(
        coords, physics_specs)
    state = initial_state_fn(jax.random.PRNGKey(0))
    modal_orography = primitive_equations.truncated_modal_orography(
        aux_features[xarray_utils.OROGRAPHY], coords)
    expected_geopotential = aux_features[xarray_utils.GEOPOTENTIAL_KEY]
    with self.subTest('dry_geopotential'):
      actual = coords.horizontal.to_nodal(
          primitive_equations.get_geopotential(
              state.temperature_variation,
              aux_features[xarray_utils.REF_TEMP_KEY],
              modal_orography,
              coords.vertical))
      np.testing.assert_allclose(actual, expected_geopotential, atol=atol)
    with self.subTest('moist_geopotential'):
      temperature = (
          aux_features[xarray_utils.REF_TEMP_KEY][:, np.newaxis, np.newaxis] +
          coords.horizontal.to_nodal(state.temperature_variation))
      specific_humidity = jnp.zeros_like(temperature)
      nodal_orography = coords.horizontal.to_nodal(modal_orography)
      actual = primitive_equations.get_geopotential_with_moisture(
          temperature, specific_humidity, nodal_orography, coords.vertical)
      np.testing.assert_allclose(actual, expected_geopotential, atol=atol)

  def testStationarySolution(self):
    """Tests that steady state is stationary for primitive equations."""
    wavenumbers = 42
    layers = 26
    dt_si = 600 * units.s
    save_every_si = 4 * units.hour
    inner_steps = int(save_every_si / dt_si)
    outer_steps = 6
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    dt = physics_specs.nondimensionalize(dt_si)
    coords = coordinate_systems.CoordinateSystem(
        horizontal=spherical_harmonic.Grid.with_wavenumbers(wavenumbers),
        vertical=sigma_coordinates.SigmaCoordinates.equidistant(layers))
    initial_state_fn, aux_features = primitive_equations_states.steady_state_jw(
        coords, physics_specs)
    ref_temps = aux_features[xarray_utils.REF_TEMP_KEY]
    modal_orography = primitive_equations.truncated_modal_orography(
        aux_features[xarray_utils.OROGRAPHY], coords)
    state = initial_state_fn()
    tracer_names = ['tracer_a', 'tracer_b']
    tracer_amplitudes = [1.5, 2.5]
    state.tracers = {
        name: primitive_equations_states.gaussian_scalar(
            coords, physics_specs, amplitude=amplitude)
        for name, amplitude in zip(tracer_names, tracer_amplitudes)}
    primitive = primitive_equations.PrimitiveEquations(
        ref_temps, modal_orography, coords, physics_specs)
    step_fn = time_integration.semi_implicit_leapfrog(primitive, dt)
    filters = (
        time_integration.exponential_leapfrog_step_filter(
            coords.horizontal, dt),
        time_integration.robert_asselin_leapfrog_filter(0.05),
    )
    step_fn = time_integration.step_with_filters(step_fn, filters)
    post_process_fn = lambda x: x[0]  # select slices of leapfrog tuple.
    trajectory_fn = time_integration.trajectory_from_step(
        step_fn, outer_steps, inner_steps, post_process_fn=post_process_fn)
    trajectory_fn = jax.jit(trajectory_fn)
    input_state = (state, state)
    _, trajectory = trajectory_fn(input_state)
    trajectory = jax.device_get(trajectory)

    def tracer_integral(tracer):
      tracer_nodal = coords.horizontal.to_nodal(tracer)
      tracer_columns = sigma_coordinates.sigma_integral(
          tracer_nodal, coords.vertical, keepdims=False)
      return coords.horizontal.integrate(tracer_columns)

    expected_tracer_sums = {
        tracer_name: tracer_integral(state.tracers[tracer_name])
        for tracer_name in tracer_names}
    for step in range(outer_steps):
      with self.subTest(f'Divergence remains close to zero, step {step}'):
        np.testing.assert_array_less(abs(trajectory.divergence[step]), 1e-3)

      with self.subTest(f'Vorticity is stationary, step {step}'):
        np.testing.assert_allclose(
            trajectory.vorticity[step], state.vorticity, atol=5e-4)

      with self.subTest(f'Temperature is stationary, step {step}'):
        np.testing.assert_allclose(
            trajectory.temperature_variation[step], state.temperature_variation,
            atol=5e-2)

      with self.subTest(f'Log surface pressure is stationary, step {step}'):
        np.testing.assert_allclose(
            trajectory.log_surface_pressure[step], state.log_surface_pressure,
            atol=5e-4)

      with self.subTest(f'Conservation of tracer, step {step}'):
        # Note: mass is not conserved by construction, but should change rather
        # slowly during smooth evolution.
        for tracer_name in tracer_names:
          actual_tracer_sum = tracer_integral(
              trajectory.tracers[tracer_name][step])
          expected_tracer_sum = expected_tracer_sums[tracer_name]
          np.testing.assert_allclose(
              actual_tracer_sum / expected_tracer_sum, 1, atol=3e-5)

  @parameterized.parameters(
      dict(coordinates=sigma_coordinates.SigmaCoordinates.equidistant(10),
           ideal_gas_constant=1),
      dict(coordinates=sigma_coordinates.SigmaCoordinates.equidistant(21),
           ideal_gas_constant=12.3),
  )
  def testGetGeopotentialWeights(self, coordinates, ideal_gas_constant):
    """Tests that the entries of geopotential weights `G` are correct."""
    g = primitive_equations.get_geopotential_weights(coordinates,
                                                     ideal_gas_constant)
    np.testing.assert_array_equal(
        [coordinates.layers, coordinates.layers], g.shape)
    alpha = primitive_equations.get_sigma_ratios(coordinates)
    for i in range(coordinates.layers):
      for j in range(coordinates.layers):

        #            𝜶[0]    𝜶[0] + 𝜶[1]    𝜶[1] + 𝜶[2]    𝜶[2] + 𝜶[3]    ᠁
        # G / R  =   0       𝜶[1]           𝜶[1] + 𝜶[2]    𝜶[2] + 𝜶[3]    ᠁
        #            0       0              𝜶[2]           𝜶[2] + 𝜶[3]    ᠁
        #            ⋮       ⋮               ⋮              ⋮              ⋱

        if i > j:
          expected_entry = 0
        elif i == j:
          expected_entry = ideal_gas_constant * alpha[j]
        else:
          expected_entry = ideal_gas_constant * (alpha[j] + alpha[j - 1])
        np.testing.assert_almost_equal(expected_entry, g[i, j],
                                       err_msg=f'Mismatch on entry {[i, j]}.')

  def testGetGeopotentialDiffBothWays(self):
    temperature = np.random.RandomState(0).randn(12, 1, 1)
    coordinates = sigma_coordinates.SigmaCoordinates.equidistant(12)
    ideal_gas_constant = 1.5
    result_matvec = primitive_equations.get_geopotential_diff(
        temperature, coordinates, ideal_gas_constant, method='dense')
    result_cumsum = primitive_equations.get_geopotential_diff(
        temperature, coordinates, ideal_gas_constant, method='sparse')
    np.testing.assert_allclose(result_matvec, result_cumsum, atol=1e-6)

  @parameterized.parameters(
      dict(coordinates=sigma_coordinates.SigmaCoordinates.equidistant(5),
           reference_temperature=np.linspace(100, 200, 5),
           heat_capacity_ratio=.5),
      dict(coordinates=sigma_coordinates.SigmaCoordinates.equidistant(23),
           reference_temperature=np.linspace(250, 300, 23),
           heat_capacity_ratio=.2857),
  )
  def testGetTemperatureImplcitWeights(
      self, coordinates, reference_temperature, heat_capacity_ratio):
    """Tests that the entries of temperature weights `H` are correct."""
    h = primitive_equations.get_temperature_implicit_weights(
        coordinates, reference_temperature, heat_capacity_ratio)
    np.testing.assert_array_equal(
        [coordinates.layers, coordinates.layers], h.shape)
    alpha = primitive_equations.get_sigma_ratios(coordinates)

    def k(r, s):
      """Computes the term denoted `K[r, s]` in the code."""
      assert r >= -1
      assert r <= coordinates.layers - 1
      assert s >= 0
      assert s <= coordinates.layers - 1

      # K[r, s] = (T[r + 1] - T[r]) / (Δ𝜎[r + 1] + Δ𝜎[r])
      #           · (P(r - s) - sum(Δ𝜎[:r + 1]))
      # K[r, s] = 0  if r < 0
      # K[r, s] = 0  when `r = coordinates.layers - 1`

      if r < 0:
        return 0
      if r == coordinates.layers - 1:
        return 0

      return (((r - s >= 0) - coordinates.layer_thickness[:r + 1].sum())
              * (reference_temperature[r + 1] - reference_temperature[r])
              / (coordinates.layer_thickness[r + 1]
                 + coordinates.layer_thickness[r]))

    for r in range(coordinates.layers):
      for s in range(coordinates.layers):

        # H[r, s] / Δ𝜎[s] = 𝜅T[r] · (P(r-s) 𝛼[r] + P(r-s-1) 𝛼[r-1]) / Δ𝜎[r]
        #           - ̇K[r, s]
        #           - K[r-1, s]

        expected_entry = coordinates.layer_thickness[s] * (
            heat_capacity_ratio * reference_temperature[r]
            * ((r - s >= 0) * alpha[r] + (r - s - 1 >= 0) * alpha[r - 1])
            / coordinates.layer_thickness[r]
            - k(r, s) - k(r - 1, s))
        np.testing.assert_almost_equal(expected_entry, h[r, s],
                                       err_msg=f'Mismatch in entry {[r, s]}.')

  @parameterized.named_parameters(
      dict(testcase_name='variable_reference_temperature',
           reference_temperature=np.linspace(100, 200, 5)),
      dict(testcase_name='constant_reference_temperature',
           reference_temperature=100*np.ones(5)),
  )
  def testGetTemperatureImplicitBothWays(self, reference_temperature):
    divergence = np.random.RandomState(0).randn(5, 1, 1)
    coordinates = sigma_coordinates.SigmaCoordinates.equidistant(5)
    result_matvec = primitive_equations.get_temperature_implicit(
        divergence, coordinates, reference_temperature, method='dense')
    result_cumsum = primitive_equations.get_temperature_implicit(
        divergence, coordinates, reference_temperature, method='sparse')
    np.testing.assert_allclose(result_matvec, result_cumsum, atol=1e-5)

  @parameterized.parameters(
      dict(wavenumbers=32, layers=4),
      dict(wavenumbers=64, layers=10),
  )
  def testPrimitiveEquationsExplicitShape(self, wavenumbers, layers):
    """Tests that output of primitive_equations_explicit has expected shape."""
    coords = coordinate_systems.CoordinateSystem(
        horizontal=spherical_harmonic.Grid.with_wavenumbers(wavenumbers),
        vertical=sigma_coordinates.SigmaCoordinates.equidistant(layers))
    reference_temperature = 300 * np.ones(layers)
    l, _ = coords.horizontal.modal_mesh
    modal_orography = np.zeros_like(l)
    vorticity = jnp.ones((layers,) + l.shape)
    divergence = jnp.ones((layers,) + l.shape)
    temperature_variation = jnp.ones((layers,) + l.shape)
    log_surface_pressure = jnp.ones((1,) + l.shape)

    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    state = primitive_equations.State(
        vorticity, divergence, temperature_variation, log_surface_pressure)
    primitive = primitive_equations.PrimitiveEquations(
        reference_temperature, modal_orography, coords, physics_specs)

    output = primitive.explicit_terms(state)
    with self.subTest('divergence shape'):
      self.assertEqual(state.divergence.shape, output.divergence.shape)
    with self.subTest('vorticity shape'):
      self.assertEqual(state.vorticity.shape, output.vorticity.shape)
    with self.subTest('temperature shape'):
      self.assertEqual(state.temperature_variation.shape,
                       output.temperature_variation.shape)
    with self.subTest('log_surface_pressure shape'):
      self.assertEqual(state.log_surface_pressure.shape,
                       output.log_surface_pressure.shape)

  @parameterized.parameters(
      dict(wavenumbers=64, layers=10),
  )
  def testPrimitiveEquationsExplicitScalesInvariance(self, wavenumbers, layers):
    """Tests that tendencies in SI units are not affected by scales."""
    default_scale = scales.DEFAULT_SCALE
    custom_scale = scales.Scale(
        scales.RADIUS / 100,
        55.3 / 2 / scales.OMEGA,
        1 * units.kilogram * 16.4,
        1 * units.degK * 3.15)
    physics_specs_a = primitive_equations.PrimitiveEquationsSpecs.from_si(
        scale=default_scale)
    grid_a = spherical_harmonic.Grid.with_wavenumbers(
        wavenumbers, radius=physics_specs_a.radius)
    physics_specs_b = primitive_equations.PrimitiveEquationsSpecs.from_si(
        scale=custom_scale)
    grid_b = spherical_harmonic.Grid.with_wavenumbers(
        wavenumbers, radius=physics_specs_b.radius)
    vertical_grid = sigma_coordinates.SigmaCoordinates.equidistant(layers)
    coords_a = coordinate_systems.CoordinateSystem(grid_a, vertical_grid)
    coords_b = coordinate_systems.CoordinateSystem(grid_b, vertical_grid)

    # defining input states using different grids and scales;
    initial_state_fn, aux_features_a = (
        primitive_equations_states.steady_state_jw(coords_a, physics_specs_a))
    modal_orography_a = primitive_equations.truncated_modal_orography(
        aux_features_a[xarray_utils.OROGRAPHY], coords_a)
    state_a = initial_state_fn()
    state_a = state_a + primitive_equations_states.baroclinic_perturbation_jw(
        coordinate_systems.CoordinateSystem(grid_a, vertical_grid),
        physics_specs_a)
    initial_state_fn, aux_features_b = (
        primitive_equations_states.steady_state_jw(coords_b, physics_specs_b))
    modal_orography_b = primitive_equations.truncated_modal_orography(
        aux_features_b[xarray_utils.OROGRAPHY], coords_b)
    state_b = initial_state_fn()
    state_b = state_b + primitive_equations_states.baroclinic_perturbation_jw(
        coordinate_systems.CoordinateSystem(grid_b, vertical_grid),
        physics_specs_b)

    # Computing tendencies using both variations.
    primitive_a = primitive_equations.PrimitiveEquations(
        aux_features_a[xarray_utils.REF_TEMP_KEY],
        modal_orography_a,
        coordinate_systems.CoordinateSystem(grid_a, vertical_grid),
        physics_specs_a)

    primitive_b = primitive_equations.PrimitiveEquations(
        aux_features_b[xarray_utils.REF_TEMP_KEY],
        modal_orography_b,
        coordinate_systems.CoordinateSystem(grid_b, vertical_grid),
        physics_specs_b)
    tendencies_a = primitive_a.explicit_terms(state_a)
    tendencies_b = primitive_b.explicit_terms(state_b)

    with self.subTest('divergence tendency'):
      divergence_a = physics_specs_a.dimensionalize(
          tendencies_a.divergence, 1 / units.s ** 2)
      divergence_b = physics_specs_b.dimensionalize(
          tendencies_b.divergence, 1 / units.s ** 2)
      np.testing.assert_allclose(
          divergence_a.magnitude, divergence_b.magnitude, atol=5e-7)
    with self.subTest('vorticity tendency'):
      vorticity_a = physics_specs_a.dimensionalize(
          tendencies_a.vorticity, 1 / units.s ** 2)
      vorticity_b = physics_specs_b.dimensionalize(
          tendencies_b.vorticity, 1 / units.s ** 2)
      np.testing.assert_allclose(
          vorticity_a.magnitude, vorticity_b.magnitude, atol=5e-7)
    with self.subTest('temperature tendency'):
      temperature_a = physics_specs_a.dimensionalize(
          tendencies_a.temperature_variation, units.degK / units.s)
      temperature_b = physics_specs_b.dimensionalize(
          tendencies_b.temperature_variation, units.degK / units.s)
      np.testing.assert_allclose(
          temperature_a.magnitude, temperature_b.magnitude, atol=5e-7)
    with self.subTest('surface pressure tendency'):
      pressure_a = physics_specs_a.dimensionalize(
          np.exp(tendencies_a.log_surface_pressure), units.pascal / units.s)
      pressure_b = physics_specs_b.dimensionalize(
          np.exp(tendencies_b.log_surface_pressure), units.pascal / units.s)
      np.testing.assert_allclose(
          pressure_a.magnitude, pressure_b.magnitude, atol=1e-7)

  @parameterized.parameters(
      dict(grid=spherical_harmonic.Grid.with_wavenumbers(16),
           vertical_grid=sigma_coordinates.SigmaCoordinates.equidistant(5),
           reference_temperature=np.linspace(100, 200, 5),
           kappa=1.4 * units.dimensionless,
           ideal_gas_constant=33 * units.J / units.kilogram / units.degK,
           step_size=.3,
           method='split',
           seed=0),
      dict(grid=spherical_harmonic.Grid.with_wavenumbers(16),
           vertical_grid=sigma_coordinates.SigmaCoordinates.equidistant(5),
           reference_temperature=np.linspace(100, 200, 5),
           kappa=1.4 * units.dimensionless,
           ideal_gas_constant=33 * units.J / units.kilogram / units.degK,
           step_size=.3,
           method='blockwise',
           seed=0),
      dict(grid=spherical_harmonic.Grid.with_wavenumbers(16),
           vertical_grid=sigma_coordinates.SigmaCoordinates.equidistant(5),
           reference_temperature=np.linspace(100, 200, 5),
           kappa=1.4 * units.dimensionless,
           ideal_gas_constant=33 * units.J / units.kilogram / units.degK,
           step_size=.3,
           method='stacked',
           seed=0),
      dict(grid=spherical_harmonic.Grid.with_wavenumbers(128),
           vertical_grid=sigma_coordinates.SigmaCoordinates.equidistant(23),
           reference_temperature=np.linspace(250, 300, 23),
           kappa=111 * units.dimensionless,
           ideal_gas_constant=1 * units.J / units.kilogram / units.degK,
           step_size=.1,
           method='split',
           seed=1),
  )
  def testPrimitiveInverse(self, vertical_grid, grid, reference_temperature,
                           kappa, ideal_gas_constant, step_size, method, seed):
    """`primitive_inverse` computes (1 - step_size · primitive_implicit)⁻¹."""
    coords = coordinate_systems.CoordinateSystem(grid, vertical_grid)
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si(
        ideal_gas_constant_si=ideal_gas_constant,
        kappa_si=kappa)
    state = random_state(coords, jax.random.PRNGKey(seed))
    l, _ = coords.horizontal.modal_mesh
    modal_orography = np.zeros_like(l)
    primitive = primitive_equations.PrimitiveEquations(
        reference_temperature, modal_orography, coords, physics_specs)
    implicit_terms = primitive.implicit_terms(state)
    primitive_equations.validate_state_shape(implicit_terms, coords)

    with self.subTest('RequiresStaticEta'):
      # Tests that inversion fails if `step_size` is not a static value.
      with self.assertRaisesRegex(TypeError, '`step_size` must be concrete'):
        jitted_inverse = jax.jit(lambda s, t: primitive.implicit_inverse(s, t))  # pylint: disable=unnecessary-lambda
        _ = jitted_inverse(state - step_size * implicit_terms, step_size)

    jitted_inverse = jax.jit(
        lambda s: primitive.implicit_inverse(s, step_size, method))
    inverted_state = jitted_inverse(state - step_size * implicit_terms)
    primitive_equations.validate_state_shape(inverted_state, coords)
    assert_states_close(state, inverted_state, atol=1e-5)

  def testEquivalenceOfPrimitiveEquationsWithAndWithoutHumidity(self):
    """Tests that primitive equations + humidity reduces to default for q=0."""
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    horizontal = spherical_harmonic.Grid.T21()
    vertical = sigma_coordinates.SigmaCoordinates.equidistant(4)
    coords = coordinate_systems.CoordinateSystem(horizontal, vertical)

    # defining input states using different grids and scales;
    initial_state_fn, aux_features = primitive_equations_states.steady_state_jw(
        coords, physics_specs)
    modal_orography = primitive_equations.truncated_modal_orography(
        aux_features[xarray_utils.OROGRAPHY], coords)
    state = initial_state_fn()
    state = state + primitive_equations_states.baroclinic_perturbation_jw(
        coords, physics_specs)
    state.tracers = {
        'specific_humidity': primitive_equations_states.gaussian_scalar(
            coords, physics_specs, amplitude=0.0)
    }
    # Computing tendencies using both variations.
    ref_temps = aux_features[xarray_utils.REF_TEMP_KEY]
    primitive_a = primitive_equations.PrimitiveEquationsWithTime(
        ref_temps, modal_orography, coords, physics_specs)
    primitive_b = primitive_equations.MoistPrimitiveEquations(
        ref_temps, modal_orography, coords, physics_specs)

    tendencies_a = primitive_a.explicit_terms(state)
    tendencies_b = primitive_b.explicit_terms(state)

    jax.tree_util.tree_map(
        lambda x, y: np.testing.assert_allclose(x, y, atol=1e-7),
        tendencies_a, tendencies_b)


class PrimitiveEquationsSpecsTest(parameterized.TestCase):

  @parameterized.parameters(
      (np.timedelta64(1, 'h'),),
      (np.timedelta64(1, 'm'),),
      (np.timedelta64(1, 's'),),
      (np.arange(5).astype('timedelta64[s]'),),
  )
  def test_timedelta_roundtrip(self, timedelta):
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    dt = physics_specs.nondimensionalize_timedelta64(timedelta)
    actual = physics_specs.dimensionalize_timedelta64(dt)
    np.testing.assert_equal(actual, timedelta)

  @parameterized.parameters(
      dict(value=1.0, expected=6856),  # rounded down from 6856.8294
      dict(value=1e-4, expected=0),  # rounded down from 0.0.68568294
  )
  def test_equivalent_rounding_behavior(self, value, expected):
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    array_value = np.array(value)
    actual_scalar = physics_specs.dimensionalize_timedelta64(value)
    acutal_array = physics_specs.dimensionalize_timedelta64(array_value)
    self.assertEqual(actual_scalar, expected)
    self.assertEqual(acutal_array, np.array(actual_scalar))


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools

from absl.testing import absltest
from absl.testing import parameterized

from dinosaur import fourier
import numpy as np


class RealFourierTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(wavenumbers=4, nodes=7),
      dict(wavenumbers=11, nodes=11),
      dict(wavenumbers=32, nodes=63),
  )
  def testBasis(self, wavenumbers, nodes):
    f = fourier.real_basis(wavenumbers, nodes)
    nodes = np.linspace(0, 2 * np.pi, nodes, endpoint=False)
    np.testing.assert_allclose(f[:, 0], 1 / np.sqrt(2 * np.pi))
    for j in range(1, wavenumbers):
      np.testing.assert_allclose(
          f[:, 2 * j - 1], np.cos(j * nodes) / np.sqrt(np.pi), atol=1e-12
      )
      np.testing.assert_allclose(
          f[:, 2 * j], np.sin(j * nodes) / np.sqrt(np.pi), atol=1e-12
      )

  @parameterized.parameters(
      dict(wavenumbers=4, seed=0),
      dict(wavenumbers=11, seed=0),
      dict(wavenumbers=32, seed=0),
  )
  def testDerivatives(self, wavenumbers, seed):
    f = np.random.RandomState(seed).normal(size=[2 * wavenumbers - 1])
    f_x = fourier.real_basis_derivative(f)
    np.testing.assert_allclose(f_x[0], 0)
    for j in range(1, wavenumbers):
      np.testing.assert_allclose(f_x[2 * j - 1], j * f[2 * j], atol=1e-12)
      np.testing.assert_allclose(f_x[2 * j], -j * f[2 * j - 1], atol=1e-12)

  @parameterized.parameters(
      dict(wavenumbers=4),
      dict(wavenumbers=16),
      dict(wavenumbers=256),
  )
  def testNormalized(self, wavenumbers):
    """Tests that the basis functions are normalized on [0, 2π]."""
    nodes = 2 * wavenumbers - 1
    f = fourier.real_basis(wavenumbers, nodes)
    _, w = fourier.quadrature_nodes(nodes)
    eye = np.eye(2 * wavenumbers - 1)
    np.testing.assert_allclose((f.T * w).dot(f), eye, atol=1e-12)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines the dinosaur module."""

import dinosaur.coordinate_systems
import dinosaur.filtering
import dinosaur.held_suarez
import dinosaur.horizontal_interpolation
import dinosaur.jax_numpy_utils
import dinosaur.layer_coordinates
import dinosaur.primitive_equations
import dinosaur.primitive_equations_states
import dinosaur.pytree_utils
import dinosaur.radiation
import dinosaur.scales
import dinosaur.shallow_water
import dinosaur.shallow_water_states
import dinosaur.sigma_coordinates
import dinosaur.spherical_harmonic
import dinosaur.time_integration
import dinosaur.typing
import dinosaur.vertical_interpolation
import dinosaur.weatherbench_utils
import dinosaur.xarray_utils

__version__ = "1.2.0"  # keep sync with pyproject.toml



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Resting and perturbed initial states for primitive equations atmosphere."""

from typing import Callable

from dinosaur import coordinate_systems
from dinosaur import filtering
from dinosaur import primitive_equations
from dinosaur import scales
from dinosaur import typing
from dinosaur import xarray_utils

import jax
import jax.numpy as jnp
import numpy as np

units = scales.units

Array = typing.Array
Quantity = typing.Quantity
QuantityOrStr = Quantity | str


def isothermal_rest_atmosphere(
    coords: coordinate_systems.CoordinateSystem,
    physics_specs: primitive_equations.PrimitiveEquationsSpecs,
    tref: QuantityOrStr = 288. * units.degK,
    p0: QuantityOrStr = 1e5 * units.pascal,
    p1: QuantityOrStr = 0. * units.pascal,
    surface_height: Quantity | None = None,
) -> tuple[Callable[..., primitive_equations.State], typing.AuxFeatures]:
  """Returns a function that generates random states and static features.

  This function implements initial states for Held-Suarez test [1].
  The returned features includes reference temperatures and flat orography.

  The choice of specifying the isothermal atmosphere with uniform T_ref is to
  avoid the Simmons-Hoskins-Burridge instability [2], also see page 601 of [3].

  References:

    [1]: Held, I. M., & Suarez, M. J. (1994). A proposal for the intercomparison
         of the dynamical cores of atmospheric general circulation models.

    [2]: Simmons, A. J., Hoskins, B. J., Burridge, D. M. (1978). Stability of
         the semi-implicit method of time integration.

    [3]: Satoh, M. (2014). Atmospheric Circulation and General Circulation
         Models, 2nd ed.

  Args:
    coords: horizontal and vertical descritization.
    physics_specs: object holding physical constants and definition of custom
      units to use for initialization of the state.
    tref: horizontal reference temperature at all altitudes.
    p0: reference surface pressure.
    p1: magnitude of surface pressure perturbation.
    surface_height: altitude above sea level. Must be an Array with units
      defined on the nodal locations form coords.horizontal.nodal_mesh.
      Expected to have Pint units assigned so that it can be nondimenionalized
      by physics_specs.

  Returns:
    state: a steady atmosphere state for the primitive equations.
    orography: an orography data in modal representation.
    reference_temperatures: a reference temperatures for all vertical levels.
  """
  lon, sin_lat = coords.horizontal.nodal_mesh
  lat = np.arcsin(sin_lat)
  tref = physics_specs.nondimensionalize(units.Quantity(tref))
  p0 = physics_specs.nondimensionalize(units.Quantity(p0))
  p1 = physics_specs.nondimensionalize(units.Quantity(p1))
  if surface_height is None:
    orography = np.zeros_like(lat)  # flat planet
  else:
    orography = np.asarray(physics_specs.nondimensionalize(surface_height))
    if orography.shape != lat.shape:
      raise ValueError(f'Expected surface_height to have shape {lat.shape}, '
                       f'got {surface_height.shape}')

  def _get_vorticity(sigma, lon, lat):
    """Computes vorticity at sigma-level as a function of lon, lat."""
    del sigma, lon  # unused.
    return jnp.zeros_like(lat)

  def _get_surface_pressure(lon, lat, rng_key):
    """Computes surface pressure as a function of lon, lat."""
    # --Get the surface pressure due to orography--
    def relative_pressure(altitude_m):
      # https://en.wikipedia.org/wiki/Atmospheric_pressure
      # pylint: disable=invalid-name
      g = 9.80665  # m/s2
      cp = 1004.68506  # J/(kg·K)
      T0 = 288.16  # K
      M = 0.02896968  # kg/mol
      R0 = 8.314462618  # J/(mol·K)
      return (1 - g * altitude_m / (cp * T0))**(cp * M / R0)
    # pylint: enable=invalid-name

    # Get nodal altitude in meters from modal orography
    # Note that this will likely be different from surface height since the
    # orography is represented with a finite number of spectral basis functions
    altitude_m = physics_specs.dimensionalize(orography, units.meter).magnitude
    surface_pressure = (p0 * np.ones(coords.surface_nodal_shape)
                        * relative_pressure(altitude_m))

    # --Add a surface pressure perturbation--
    keys = jax.random.split(rng_key, 2)
    lon0 = jax.random.uniform(keys[1], minval=np.pi / 2, maxval=3 * np.pi / 2)
    lat0 = jax.random.uniform(keys[0], minval=-np.pi / 4, maxval=np.pi / 4)
    stddev = np.pi / 20  # std deviation in lon, lat
    k = 4  # wavenumber in lon
    perturbation = (jnp.exp(-(lon - lon0)**2 / (2 * stddev**2)) *
                    jnp.exp(-(lat - lat0)**2 / (2 * stddev**2)) *
                    jnp.sin(k * (lon - lon0)))
    return surface_pressure + p1 * perturbation

  def random_state_fn(rng_key: jnp.ndarray) -> primitive_equations.State:
    nodal_vorticity = jnp.stack(
        [_get_vorticity(sigma, lon, lat) for sigma in coords.vertical.centers])
    modal_vorticity = coords.horizontal.to_modal(nodal_vorticity)
    nodal_surface_pressure = _get_surface_pressure(lon, lat, rng_key)
    return primitive_equations.State(
        vorticity=modal_vorticity,
        divergence=jnp.zeros_like(modal_vorticity),
        temperature_variation=jnp.zeros_like(modal_vorticity),
        log_surface_pressure=(
            coords.horizontal.to_modal(jnp.log(nodal_surface_pressure))),
        )

  aux_features = {
      xarray_utils.OROGRAPHY: orography,
      xarray_utils.REF_TEMP_KEY: np.full((coords.vertical.layers,), tref)
  }
  return random_state_fn, aux_features


def isothermal_rest_atmosphere_with_orography_path(
    coords: coordinate_systems.CoordinateSystem,
    physics_specs: primitive_equations.PrimitiveEquationsSpecs,
    path_to_orography_data: str,
    tref: QuantityOrStr = 288. * units.degK,
    p0: QuantityOrStr = 1e5 * units.pascal,
    p1: QuantityOrStr = 0. * units.pascal,
) -> tuple[Callable[..., primitive_equations.State], typing.AuxFeatures]:
  """Wrapper around `isothermal_rest_atmosphere` that loads orography data."""
  ds = xarray_utils.open_dataset(path_to_orography_data)
  input_coords = xarray_utils.coordinate_system_from_dataset(
      ds, spmd_mesh=coords.spmd_mesh)
  filter_fns = [filtering.exponential_filter(coords.horizontal)]
  orography_key = xarray_utils.OROGRAPHY
  nodal_orography = ds[orography_key].transpose('lon', 'lat').values
  orography = primitive_equations.filtered_modal_orography(
      nodal_orography, coords, input_coords, filter_fns)
  nodal_orography_filtered = coords.horizontal.to_nodal(orography)
  nodal_orography_filtered = np.asarray(nodal_orography_filtered)
  return isothermal_rest_atmosphere(
      coords=coords, physics_specs=physics_specs, tref=tref, p0=p0, p1=p1,
      surface_height=(nodal_orography_filtered * units.meters))


def steady_state_jw(
    coords: coordinate_systems.CoordinateSystem,
    physics_specs: primitive_equations.PrimitiveEquationsSpecs,
    u0: Quantity = 35. * units.m / units.s,
    p0: Quantity = 1e5 * units.pascal,
    t0: Quantity = 288. * units.degK,
    delta_t: Quantity = 4.8e5 * units.degK,
    gamma: Quantity = 0.005 * units.degK / units.m,
    sigma_tropo: float = 0.2,
    sigma0: float = 0.252,
) -> tuple[Callable[..., primitive_equations.State], typing.AuxFeatures]:
  """Returns a function that generates steady state and static features.

  This steady state is based on the baroclinic wave test case paper by
  Jablonowski and Williamson [1]. It is provided in a closed form and should
  remain stationary for up-to 30 simulation days. An additional perturbation
  `baroclinic_perturbation_jw` can be introduced on-top of the steady state to
  trigger a baroclinic instability. Note that to achieve stability for 30 days
  we currently need to enable 64bit precision.

  References:

    [1]: Jablonowski, C., & Williamson, D. L. (2006). A baroclinic instability
         test case for atmospheric model dynamical cores.

  Args:
    coords: horizontal and vertical descritization.
    physics_specs: object holding physical constants and definition of custom
      units to use for initialization of the state.
    u0: velocity scale of the zonal jets in the steady state.
    p0: reference pressure of the steady state.
    t0: horizontal mean temperature at the surface.
    delta_t: empirical temperature difference.
    gamma: temperature lapse rate.
    sigma_tropo: tropopause level in sigma coordinates.
    sigma0: a sigma constant used to define sigma_nu in specification of the
      initial state.

  Returns:
    state: a steady atmosphere state for the primitive equations.
    orography: an orography data in modal representation.
    reference_temperatures: a reference temperatures for all vertical levels.
  """
  u0 = physics_specs.nondimensionalize(u0)
  t0 = physics_specs.nondimensionalize(t0)
  delta_t = physics_specs.nondimensionalize(delta_t)
  p0 = physics_specs.nondimensionalize(p0)
  gamma = physics_specs.nondimensionalize(gamma)
  a = physics_specs.radius
  g = physics_specs.g
  r_gas = physics_specs.R
  omega = physics_specs.angular_velocity

  def _get_reference_temperature(sigma):
    """Computes reference temperature for a given sigma level."""
    top_mean_t = t0 * sigma ** (r_gas * gamma / g)
    if sigma < sigma_tropo:
      return top_mean_t + delta_t * (sigma_tropo - sigma) ** 5
    else:
      return top_mean_t

  def _get_reference_geopotential(sigma):
    """Computes reference geopotential for a given sigma level."""
    top_mean_potential = (t0 * g / gamma) * (1 - sigma ** (r_gas * gamma / g))
    if sigma < sigma_tropo:
      return top_mean_potential - r_gas * delta_t * (
          (np.log(sigma / sigma_tropo) + 137 / 60) * sigma_tropo ** 5 -
          5 * sigma * sigma_tropo ** 4 + 5 * (sigma ** 2) * (sigma_tropo ** 3) -
          (10 / 3) * (sigma_tropo ** 2) * sigma ** 3 +
          (5 / 4) * sigma_tropo * sigma ** 4 - (sigma ** 5) / 5
      )
    else:
      return top_mean_potential

  def _get_geopotential(lat, lon, sigma):
    """Computes geopotential at sigma-level as a function of lat."""
    del lon  # unused.
    sigma_nu = (sigma - sigma0) * np.pi / 2
    return _get_reference_geopotential(sigma) + u0 * np.cos(sigma_nu) ** 1.5 * (
        ((-2 * np.sin(lat) ** 6 * (np.cos(lat) ** 2 + 1 / 3) + 10 / 63) *
         u0 * np.cos(sigma_nu) ** (3 / 2)) +
        ((1.6 * (np.cos(lat) ** 3) * (np.sin(lat) ** 2 + 2 / 3) - np.pi / 4) *
         a * omega)
    )

  def _get_temperature_variation(lat, lon, sigma):
    """Computes temperature variation at sigma-level as a function of lat."""
    del lon  # unused.
    sigma_nu = (sigma - sigma0) * np.pi / 2
    cos_𝜎ν = np.cos(sigma_nu)  # pylint: disable=invalid-name
    sin_𝜎ν = np.sin(sigma_nu)  # pylint: disable=invalid-name
    return 0.75 * (sigma * np.pi * u0 / r_gas) * sin_𝜎ν * np.sqrt(cos_𝜎ν) * (
        ((-2 * (np.cos(lat) ** 2 + 1 / 3) * np.sin(lat) ** 6 + 10 / 63) *
         2 * u0 * cos_𝜎ν ** (3 / 2)) +
        ((1.6 * (np.cos(lat) ** 3) * (np.sin(lat) ** 2 + 2 / 3) - np.pi / 4) *
         a * omega)
    )

  def _get_vorticity(lat, lon, sigma):
    """Computes vorticity at sigma-level as a function of lat."""
    del lon  # unused.
    sigma_nu = (sigma - sigma0) * np.pi / 2
    return ((-4 * u0 / a) * (np.cos(sigma_nu) ** (3 / 2)) *
            np.sin(lat) * np.cos(lat) * (2 - 5 * np.sin(lat) ** 2))

  def _get_surface_pressure(lat, lon,):
    """Computes surface pressure as a function of lat."""
    del lon  # unused.
    return p0 * np.ones(lat.shape)[np.newaxis, ...]

  lon, sin_lat = coords.horizontal.nodal_mesh
  lat = np.arcsin(sin_lat)
  def initial_state_fn(
      rng_key: jnp.ndarray | None = None
  ) -> primitive_equations.State:
    del rng_key  # unused.
    nodal_vorticity = np.stack(
        [_get_vorticity(lat, lon, sigma) for sigma in coords.vertical.centers])
    modal_vorticity = coords.horizontal.to_modal(nodal_vorticity)
    nodal_temperature_variation = np.stack(
        [_get_temperature_variation(lat, lon, sigma)
         for sigma in coords.vertical.centers])
    # shift in log_surf_pressure due to custom units doesn't affect dynamics.
    log_nodal_surface_pressure = np.log(_get_surface_pressure(lat, lon))
    state = primitive_equations.State(
        vorticity=modal_vorticity,
        divergence=np.zeros_like(modal_vorticity),
        temperature_variation=coords.horizontal.to_modal(
            nodal_temperature_variation),
        log_surface_pressure=coords.horizontal.to_modal(
            log_nodal_surface_pressure))
    return state

  orography = _get_geopotential(lat, lon, 1.) / g
  geopotential = np.stack(
      [_get_geopotential(lat, lon, sigma) for sigma in coords.vertical.centers])
  reference_temperatures = np.stack(
      [_get_reference_temperature(sigma) for sigma in coords.vertical.centers])
  aux_features = {
      xarray_utils.GEOPOTENTIAL_KEY: geopotential,
      xarray_utils.OROGRAPHY: orography,
      xarray_utils.REF_TEMP_KEY: reference_temperatures,
  }
  return initial_state_fn, aux_features


def baroclinic_perturbation_jw(
    coords: coordinate_systems.CoordinateSystem,
    physics_specs: primitive_equations.PrimitiveEquationsSpecs,
    u_perturb: Quantity = 1. * units.m / units.s,
    lon_location: Quantity = np.pi / 9,
    lat_location: Quantity = 2 * np.pi / 9,
    perturbation_radius: Quantity = 0.1,
) -> primitive_equations.State:
  """Returns perturbation that triggers baroclinic instability.

  The perturbation to the state returned by this function triggers baroclinic
  instability in the rest state produced by `steady_state_jw` and is taken from
  the baroclinic wave test case paper by Jablonowski and Williamson [1].

  References:

    [1]: Jablonowski, C., & Williamson, D. L. (2006). A baroclinic instability
         test case for atmospheric model dynamical cores.

  Args:
    coords: horizontal and vertical descritization.
    physics_specs: object holding physical constants and definition of custom
      units to use for initialization of the state.
    u_perturb: velocity scale of the perturbation.
    lon_location: location of the perturbation along longitude.
    lat_location: location of the perturbation along latitude.
    perturbation_radius: ratio of the spatial scale of the perturbation to the
      radius of the system.

  Returns:
    a perturbation to the atmosphere state that triggers baroclinic instability.
  """
  u_p = physics_specs.nondimensionalize(u_perturb)
  a = physics_specs.radius

  def _get_vorticity_perturbation(lat, lon, sigma):
    del sigma  # unused.
    x = (np.sin(lat_location) * np.sin(lat) +
         np.cos(lat_location) * np.cos(lat) * np.cos(lon - lon_location))
    r = a * np.arccos(x)
    R = a * perturbation_radius  # pylint: disable=invalid-name
    return (u_p / a) * np.exp(-(r / R) ** 2) * (
        np.tan(lat) - (2 * ((a / R) ** 2) * np.arccos(x)) *
        (np.sin(lat_location) * np.cos(lat) -
         np.cos(lat_location) * np.sin(lat) * np.cos(lon - lon_location)) /
        (np.sqrt(1 - x ** 2)))

  def _get_divergence_perturbation(lat, lon, sigma):
    del sigma  # unused.
    x = (np.sin(lat_location) * np.sin(lat) +
         np.cos(lat_location) * np.cos(lat) * np.cos(lon - lon_location))
    r = a * np.arccos(x)
    R = a * perturbation_radius  # pylint: disable=invalid-name
    return (-2 * u_p * a / (R ** 2)) * np.exp(-(r / R) ** 2) * np.arccos(x) * (
        (np.cos(lat_location) * np.sin(lon - lon_location)) /
        (np.sqrt(1 - x ** 2)))

  lon, sin_lat = coords.horizontal.nodal_mesh
  lat = np.arcsin(sin_lat)
  nodal_vorticity = np.stack(
      [_get_vorticity_perturbation(lat, lon, sigma)
       for sigma in coords.vertical.centers])
  nodal_divergence = np.stack(
      [_get_divergence_perturbation(lat, lon, sigma)
       for sigma in coords.vertical.centers])
  modal_vorticity = coords.horizontal.to_modal(nodal_vorticity)
  modal_divergence = coords.horizontal.to_modal(nodal_divergence)
  state = primitive_equations.State(
      vorticity=modal_vorticity,
      divergence=modal_divergence,
      temperature_variation=np.zeros_like(modal_vorticity),
      log_surface_pressure=np.zeros_like(modal_vorticity[:1, ...]),)
  return state


def gaussian_scalar(
    coords: coordinate_systems.CoordinateSystem,
    physics_specs: primitive_equations.PrimitiveEquationsSpecs,
    lon_location: float = np.pi / 9,
    lat_location: float = 2 * np.pi / 9,
    perturbation_radius: float = 0.2,
    amplitude: float = 1.
) -> Array:
  """Returns a scalar gaussian field in modal representation.

  Args:
    coords: horizontal and vertical descritization.
    physics_specs: object holding physical constants and definition of custom
      units to use for initialization of the state.
    lon_location: location of the field along longitude.
    lat_location: location of the field along latitude.
    perturbation_radius: ratio of the spatial scale of the field to the radius.
    amplitude: maximum amplitude of the field.

  Returns:
    Array holding modal values of the field.
  """
  a = physics_specs.radius

  def _get_field_values(lat, lon, sigma):
    del sigma  # unused.
    x = (np.sin(lat_location) * np.sin(lat) +
         np.cos(lat_location) * np.cos(lat) * np.cos(lon - lon_location))
    r = a * np.arccos(x)
    R = a * perturbation_radius  # pylint: disable=invalid-name
    return amplitude * np.exp(-(r / R) ** 2)

  lon, sin_lat = coords.horizontal.nodal_mesh
  lat = np.arcsin(sin_lat)
  return coords.horizontal.to_modal(
      np.stack([_get_field_values(lat, lon, sigma)
                for sigma in coords.vertical.centers]))



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for shallow_water.

The validation tests below are based on

"A standard test set for numerical approximations to the shallow water equations
in spherical geometry"
David L.Williamson, John B.Drake, James J.Hack, Rüdiger Jakob,
Paul N.Swarztrauber
https://doi.org/10.1016/S0021-9991(05)80016-6

At present, the tests cover only test case 2 from the paper, "Steady State
Nonlinear Zonal Geostrophic Flow." We plan to add additional test cases as we
build out the feature set of the solver.
"""
import unittest

from absl.testing import absltest
from absl.testing import parameterized

from dinosaur import associated_legendre
from dinosaur import coordinate_systems
from dinosaur import layer_coordinates
from dinosaur import scales
from dinosaur import shallow_water
from dinosaur import shallow_water_states
from dinosaur import spherical_harmonic

import jax
import jax.numpy as jnp
import numpy as np

units = scales.units
jax.config.parse_flags_with_absl()


def _tpu_or_gpu_available():
  return jax.devices()[0].platform != 'cpu'


def assert_states_close(state0, state1, **kwargs):
  for field in state0.fields:
    np.testing.assert_allclose(getattr(state0, field.name),
                               getattr(state1, field.name),
                               err_msg=f'Mismatch in {field}:',
                               **kwargs)


def _get_mountain(grid, height, physics_specs):
  """Returns the orography for a mountain at (3π / 2, π / 6)."""
  mountain_geopotential = (
      physics_specs.nondimensionalize(height) * physics_specs.g)
  center_lon = 3 * np.pi / 2
  center_lat = np.pi / 6
  lon, sin_lat = grid.nodal_mesh
  lat = np.arcsin(sin_lat)
  r = np.pi / 9
  d = np.sqrt(np.minimum(r**2, (lon - center_lon)**2 + (lat - center_lat)**2))
  mountain_nodal = mountain_geopotential * (1 - d / r)
  mountain = grid.to_modal(mountain_nodal)
  return mountain


def _get_geopotential(grid, max_velocity, thickness, layers, physics_specs):
  """Mean geopotential and fluctuation corresponding to velocity u0 · sin 𝜃."""
  _, sin_lat = grid.nodal_mesh
  gh0 = physics_specs.nondimensionalize(thickness) * physics_specs.g
  max_v = physics_specs.nondimensionalize(max_velocity)
  total_geopotential = gh0 - (
      physics_specs.radius * physics_specs.angular_velocity * max_v +
      max_v ** 2 / 2) * sin_lat ** 2
  geopotential = jnp.stack([total_geopotential / layers] * layers)

  _, w = associated_legendre.gauss_legendre_nodes(grid.latitude_nodes)
  mean_geopotential = (
      (geopotential * w).sum((-1, -2)) / w.sum() / grid.longitude_nodes)
  delta_geopotential = grid.to_modal(
      geopotential - mean_geopotential[..., jnp.newaxis, jnp.newaxis])
  return mean_geopotential, delta_geopotential


def _compute_mass(grid, potential, mean_geopotential, density):
  """Computes the total mass in arbitrary units."""
  layers = density.shape[0]
  _, w = associated_legendre.gauss_legendre_nodes(grid.latitude_nodes)
  total_potential = (grid.to_nodal(potential) +
                     jnp.expand_dims(mean_geopotential, (1, 2)))
  volume = ((total_potential * w).sum((-1, -2)) /
            w.sum() / grid.longitude_nodes / layers)
  return (volume * density).sum(-1)


class ShallowWaterTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(wavenumbers=64,
           layers=1,
           velocity_function=lambda lat: np.cos(3 * lat) / 5,
           dt=1e-3,
           density_ratio=.9,
           mean_potential=1/10,
           inner_steps=1000,
           outer_steps=10),
      dict(wavenumbers=64,
           layers=4,
           velocity_function=(lambda lat: np.cos(lat) / 5),
           dt=1e-3,
           density_ratio=.9,
           mean_potential=1/10,
           inner_steps=1000,
           outer_steps=10),
      dict(wavenumbers=128,
           layers=2,
           velocity_function=(lambda lat: np.cos(lat) ** 2 / 5),
           dt=1e-4,
           density_ratio=.9,
           mean_potential=1/10,
           inner_steps=1000,
           outer_steps=10),
  )
  def testSteadyStateGeostrophicFlow(self, wavenumbers, layers,
                                     velocity_function, dt, density_ratio,
                                     mean_potential, inner_steps, outer_steps):
    """Tests steady state zonal geostrophic flow."""

    if not _tpu_or_gpu_available():
      # TODO(shoyer): speed up these tests, fast enough to include in CI on
      # GitHub!
      raise unittest.SkipTest('test is too slow to run on CPU')

    grid = spherical_harmonic.Grid.with_wavenumbers(wavenumbers)
    vertical_grid = layer_coordinates.LayerCoordinates(layers)
    coords = coordinate_systems.CoordinateSystem(grid, vertical_grid)
    density = np.array([density_ratio ** n for n in range(layers)][::-1])
    physics_specs = shallow_water.ShallowWaterSpecs.from_si(
        density * scales.WATER_DENSITY)
    mean_potential = np.ones(layers) * mean_potential
    orography = None  # no orography in the geostrophic flow test case.

    # Set up time integration of the shallow water equations.
    default_filters = shallow_water.default_filters(grid, dt)
    trajectory_fn = shallow_water.shallow_water_leapfrog_trajectory(
        coords, dt, physics_specs, inner_steps, outer_steps, mean_potential,
        orography, default_filters)

    # Constructs steady state from a zonal velocity field.
    lat = np.arccos(grid.cos_lat)
    velocity = jnp.stack([velocity_function(lat)] * layers)
    state_0 = state_1 = shallow_water_states.multi_layer(
        velocity, density, coords)
    initial_state = (state_0, state_1)

    # Quantities that will be used to compute relative errors.
    init_potential = grid.to_nodal(state_0.potential)
    init_potential_l2 = np.sqrt(np.square(init_potential).sum())

    # Compute the potentials at several time steps and compare them to the
    # initial potential
    # TODO(jamieas): we need more principled expectations for the deviation
    # from reference values.
    _, trajectory = trajectory_fn(initial_state)
    potentials = grid.to_nodal(trajectory.potential)
    _, w = associated_legendre.gauss_legendre_nodes(grid.latitude_nodes)

    for j, potential in enumerate(potentials):
      step = (j + 1) * inner_steps
      with self.subTest(f'Mean potential conservation, step {step}'):
        # The mean of the fluctuations in geopotential should be zero.
        mean_potential = np.mean(w * potential / np.sum(w), axis=(-1, -2))
        np.testing.assert_array_less(np.abs(mean_potential), 1e-8)

      with self.subTest(f'Steady state L2 error, step {step}'):
        # The geopotential should stay constant over time.
        l2_error = np.sqrt(
            np.square(potential - init_potential).sum()) / init_potential_l2
        self.assertLess(l2_error, 1e-5)

  @parameterized.parameters(
      dict(wavenumbers=128,
           layers=1,
           density_ratio=.9,
           max_velocity=20 * units.meter / units.second,
           mountain_height=0 * units.meter,
           atmosphere_thickness=5960 * units.meter,
           total_time=15 * units.day,
           save_every=6 * units.hour,
           dt=60 * units.second),
      dict(wavenumbers=128,
           layers=1,
           density_ratio=.9,
           max_velocity=20 * units.meter / units.second,
           mountain_height=2000 * units.meter,
           atmosphere_thickness=5960 * units.meter,
           total_time=15 * units.day,
           save_every=6 * units.hour,
           dt=60 * units.second),
      dict(wavenumbers=64,
           layers=3,
           density_ratio=.9,
           max_velocity=25 * units.meter / units.second,
           mountain_height=3000 * units.meter,
           atmosphere_thickness=7000 * units.meter,
           total_time=15 * units.day,
           save_every=6 * units.hour,
           dt=60 * units.second),
  )
  def testFlowOverAMountainMassConservation(
      self, wavenumbers, layers, density_ratio, max_velocity, mountain_height,
      atmosphere_thickness, total_time, save_every, dt):
    """Tests that mass is conserved for a flow over a mountain."""

    if not _tpu_or_gpu_available():
      raise unittest.SkipTest('test is too slow to run on CPU')

    # This test is based on Test Case 5 from
    #  "A standard test set for numerical approximations to the shallow water
    #  equations in spherical geometry"
    #  David L.Williamson, John B.Drake, James J.Hack, Rüdiger Jakob,
    #  Paul N.Swarztrauber
    #  https://doi.org/10.1016/S0021-9991(05)80016-6

    grid = spherical_harmonic.Grid.with_wavenumbers(wavenumbers)
    density = np.array([density_ratio ** n for n in range(layers)][::-1])
    vertical_grid = layer_coordinates.LayerCoordinates(layers)
    coords = coordinate_systems.CoordinateSystem(grid, vertical_grid)
    physics_specs = shallow_water.ShallowWaterSpecs.from_si(
        density * scales.WATER_DENSITY)

    # Construct initial state.
    max_v = physics_specs.nondimensionalize(max_velocity)
    u_nodal = jnp.array(
        [[max_v * grid.cos_lat] * grid.longitude_nodes] * layers)
    v_nodal = jnp.zeros_like(u_nodal)
    cos_lat_velocity = grid.to_modal(
        jnp.stack([u_nodal, v_nodal]) / grid.cos_lat)
    vorticity = grid.curl_cos_lat(cos_lat_velocity)

    # Orography consists of a single mountain.
    orography = _get_mountain(grid, mountain_height, physics_specs)

    mean_potential, delta_potential = _get_geopotential(
        grid, max_velocity, atmosphere_thickness, layers, physics_specs)

    divergence = jnp.zeros_like(delta_potential)

    state_0 = state_1 = shallow_water.State(
        vorticity=vorticity,
        divergence=divergence,
        potential=delta_potential - orography / layers)
    initial_state = (state_0, state_1)

    # Set up time stepping.
    total_time = physics_specs.nondimensionalize(total_time)
    save_every = physics_specs.nondimensionalize(save_every)
    dt = physics_specs.nondimensionalize(dt)
    inner_steps = int(save_every / dt)
    outer_steps = int(total_time / save_every)

    # Set up time integration of the shallow water equations.
    default_filters = shallow_water.default_filters(grid, dt)
    trajectory_fn = shallow_water.shallow_water_leapfrog_trajectory(
        coords, dt, physics_specs, inner_steps, outer_steps, mean_potential,
        orography, default_filters)

    # Perform integration and check conservation of mass.
    _, trajectory = trajectory_fn(initial_state)
    initial_mass = _compute_mass(
        grid, state_0.potential, mean_potential, density)
    masses = _compute_mass(
        grid, trajectory.potential, mean_potential, density)
    np.testing.assert_allclose(masses, initial_mass, rtol=1e-6)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Associated Legendre function evaluation, and derivative coefficients."""

import functools

import numpy as np
import scipy.special as sps


def _evaluate_rhombus(n_l: int,
                      n_m: int,
                      x: np.ndarray,
                      truncation='rhombus') -> np.ndarray:
  """Associated Legendre Function (rhombus indexing).

  Evaluates the associated Legendre functions on the nodes `x`.

  Arguments:
    n_l: the maximum (exclusive) degree (often denoted `k` or `l`) of the
      computed polynomials.
    n_m: the maximum (exclusive) order (often denoted `m`) of the computed
      polynomials.
    x: vector of nodes in the range [-1, 1].
    truncation: 'rhombus' or 'triangle'. See below.

  Returns:
    Array p of shape (n_l, n_m, len(x)) such that

      p[k, m, i] = cₖₘ Pᵐₘ₊ₖ(xᵢ)

    for indices in range, and 0 otherwise. By default (truncation == 'rhombus'),
    all indices are in range. If truncation == 'triangle', then p[k, m, i] is
    nonzero only when l = m + k < n_l .

    The normalization constants cₖₘ are chosen such that each basis function
    has unit L²([-1, 1]) norm.
  """
  y = np.sqrt(1 - x * x)
  p = np.zeros((n_l, n_m, len(x)))
  p[0, 0] = p[0, 0] + 1 / np.sqrt(2)
  for m in range(1, n_m):
    p[0, m] = -np.sqrt(1 + 1 / (2 * m)) * y * p[0, m - 1]
  m_max = n_m
  for k in range(1, n_l):
    if truncation == 'triangle':
      m_max = min(n_m, n_l - k)
    m = np.arange(m_max).reshape((-1, 1))
    m2 = np.square(m)
    mk2 = np.square(m + k)
    mkp2 = np.square(m + k - 1)
    a = np.sqrt((4 * mk2 - 1) / (mk2 - m2))
    b = np.sqrt((mkp2 - m2) / (4 * mkp2 - 1))
    p[k, :m_max] = a * (x * p[k - 1, :m_max] - b * p[k - 2, :m_max])
  return p


def evaluate(n_m: int,
             n_l: int,
             x: np.ndarray) -> np.ndarray:
  """Associated Legendre Function.

  Evaluates the associated Legendre functions on the nodes x.

  Arguments:
    n_m: number of m (azimuthal wavenumber) modes. Must satisfy `n_m <= n_l`
    n_l: number of l (total wavenumber) modes.
    x: vector of nodes.

  Returns:
    Array p of shape (n_m, len(x), n_l) such that

      p[m, i, l] = cₗₘ Pᵐₗ(xᵢ) .

    Note p[m, i, l] = 0 for l < m. The normalization constants cₗₘ are chosen
    such that each basis function has unit L²([-1, 1]) norm.

  Raises:
    ValueError: if `n_m > n_l`.
  """
  if n_m > n_l:
    raise ValueError(f'Expected n_m <= n_l; got n_m = {n_m} and n_l = {n_l}.')
  r = np.transpose(
      _evaluate_rhombus(n_l=n_l, n_m=n_m, x=x, truncation='triangle'),
      (1, 2, 0))
  p = np.zeros((n_m, len(x), n_l))
  for m in range(n_m):
    p[m, :, m:n_l] = r[m, :, 0:n_l - m]
  return p


def gauss_legendre_nodes(n: int) -> tuple[np.ndarray, np.ndarray]:
  """Returns nodes and weights for Gauss-Legendre quadrature.

  Args:
    n: the number of nodes and weights to return.

  Returns:
    A pair (nodes, weights); values and weights to use for Gauss-Legendre
    quadrature, with dtype `np.float64`.
  """
  return sps.roots_legendre(n)


@functools.lru_cache(maxsize=128)
def equiangular_nodes(n: int) -> tuple[np.ndarray, np.ndarray]:
  """Returns equally spaced nodes and associated weights.

  The nodes can be interpreted as the midpoints of `n` equal sized segments. So,
  for `n = 3`, we would get `x = sin(y)` where
  `y = [-π / 3, 0, π / 3]`.

  Args:
    n: the number of nodes and weights to return.

  Returns:
    A pair (nodes, weights); values and weights to use for quadrature, with
    dtype `np.float64`.
  """
  spacing = np.pi / n
  theta = np.linspace(-np.pi / 2 + spacing / 2,
                      np.pi / 2 - spacing / 2,
                      n)
  x = np.sin(theta)
  w = _compute_weights(x)
  return x, w


@functools.lru_cache(maxsize=128)
def equiangular_nodes_with_poles(n: int) -> tuple[np.ndarray, np.ndarray]:
  """Returns equally spaced nodes and associated weights.

  The nodes equally spaced sample points between [-π/2, π/2] so for
  for `n = 3`, we would get `x = sin(y)` where `y = [-π / 2, 0, π / 2]`.

  Args:
    n: the number of nodes and weights to return.

  Returns:
    A pair (nodes, weights); values and weights to use for quadrature, with
    dtype `np.float64`.
  """
  theta = np.linspace(-np.pi / 2, np.pi / 2, n)
  x = np.sin(theta)
  w = _compute_weights(x)
  return x, w


def _compute_weights(x: np.ndarray) -> np.ndarray:
  """Computes weights for a vector of `x`."""
  # Letting Pₖ be the Legendre polynomial of degree k, we note that
  # ∫ Pₖ = 0 for all k > 0. Therefore, our weights `w` should satisfy
  # `w.dot(Pₖ(x)) = 𝛿(k, 0)`. So, we can find weights `w` using a linear solve.

  # The matrix `legendre` has `legendre[k, j] = Pₖ(x[j])`.
  legendre = evaluate(n_m=1, n_l=x.shape[0], x=x)[0].T
  z = np.zeros_like(x)
  z[0] = 1
  w = np.linalg.solve(legendre, z)

  # Since `x` spans the interval [-1, 1], the weights should sum to 2 so that we
  # get the right result when integrating `f(x) = 1` on the interval [-1, 1].
  return w / w.sum() * 2



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A vertical coordinate system based on normalized pressure.

See https://en.wikipedia.org/wiki/Sigma_coordinate_system
"""

from __future__ import annotations

import dataclasses
import functools
from typing import Sequence

from dinosaur import jax_numpy_utils
from dinosaur import typing

import jax
from jax import lax
import jax.numpy as jnp
import numpy as np


Array = typing.Array

# All `einsum`s should be done at highest available precision.
einsum = functools.partial(jnp.einsum, precision=lax.Precision.HIGHEST)


def _slice_shape_along_axis(
    x: np.ndarray,
    axis: int,
    slice_width: int = 1,
) -> tuple[int, ...]:
  """Returns a shape of `x` sliced along `axis` with width `slice_width`."""
  x_shape = list(x.shape)
  x_shape[axis] = slice_width
  return tuple(x_shape)


# TODO(dkochkov) Consider renaming `SigmaCoordinates` to `Grid` or `SigmaGrid`.
@dataclasses.dataclass(frozen=True)
class SigmaCoordinates:
  """A description of a discrete vertical coordinate system.

  Layers are indexed from the "top" of the atmosphere (𝜎 = 0) to the surface of
  the earth (𝜎 = 1).

  Attributes:
    boundaries: sigma values of the boundaries of horizontal layers. Must be an
      increasing array of values beginning with zero and ending with one. For
      `n` layers, `boundaries` has length `n + 1`.
    internal_boundaries: sigma values of the boundaries _between_ layers. For
      `n` layers, `internal_boundaries` has length `n - 1`.
    centers: sigma values of the centers of horizontal layers. For `n` layers,
      `centers` has length `n`
    layer_thickness: the thickness of each layer in sigma coordinates. For `n`
      layers, `layer_thickness` has length `n`.
    center_to_center: the distances between the centers of each layer in sigma
      coordinates. For `n` layers, `center_to_center` has length `n - 1`.
    layers: the number of layers.
  """

  boundaries: np.ndarray

  def __init__(self, boundaries: Sequence[float] | np.ndarray):
    object.__setattr__(self, 'boundaries', np.asarray(boundaries))
    if not (
        np.isclose(self.boundaries[0], 0) and np.isclose(self.boundaries[-1], 1)
    ):
      raise ValueError(
          'Expected boundaries[0] = 0, boundaries[-1] = 1, '
          f'got boundaries = {self.boundaries}'
      )
    if not all(np.diff(self.boundaries) > 0):
      raise ValueError(
          'Expected `boundaries` to be monotonically increasing, '
          f'got boundaries = {self.boundaries}'
      )

  @property
  def internal_boundaries(self) -> np.ndarray:
    return self.boundaries[1:-1]

  @property
  def centers(self) -> np.ndarray:
    return (self.boundaries[1:] + self.boundaries[:-1]) / 2

  @property
  def layer_thickness(self) -> np.ndarray:
    return np.diff(self.boundaries)

  @property
  def center_to_center(self) -> np.ndarray:
    return np.diff(self.centers)

  @property
  def layers(self) -> int:
    return len(self.boundaries) - 1

  @classmethod
  def equidistant(cls, layers: int) -> SigmaCoordinates:
    boundaries = np.linspace(0, 1, layers + 1)
    return cls(boundaries)

  def asdict(self):
    return {k: v.tolist() for k, v in dataclasses.asdict(self).items()}

  def __hash__(self):
    return hash(tuple(self.centers.tolist()))

  def __eq__(self, other):
    return isinstance(other, SigmaCoordinates) and np.array_equal(
        self.centers, other.centers
    )


# For consistency with commonly accepted notation, we use Greek letters within
# some of the functions below.
# pylint: disable=invalid-name


@jax.named_call
def centered_difference(
    x: np.ndarray, coordinates: SigmaCoordinates, axis: int = -3
) -> np.ndarray:
  """Derivative of `x` with respect to `sigma` along specified `axis`.

  The derivative is approximated as

  (∂x / ∂𝜎)[n + ½] ≈ (x[n + 1] - x[n]) / (𝜎[n + 1] - 𝜎[n])

  So, the derivatives will be located on the 'boundaries' between layers and
  will consist of one fewer values than `x`.

  Args:
    x: an array of values with `x.shape[axis] == coordinates.layers`. These
      values are "located" at `coordinates` layer centers.
    coordinates: a `SigmaCoordinates` object describing the vertical coordinates
      to differentiate against. Must satisfy `coordinates.layers ==
      x.shape[axis]`.
    axis: the axis along which to approximate the derivative. Defaults to -3
      since this is the axis we typically use to index layers.

  Returns:
    An array `x_sigma` with `x_sigma.shape[axis] == x.shape[axis] - 1`,
    containing approximate values of the derivative at of `x` at
    `coordinates.internal_boundaries`.
  """
  if coordinates.layers != x.shape[axis]:
    raise ValueError(
        '`x.shape[axis]` must be equal to `coordinates.layers`; '
        f'got {x.shape[axis]} and {coordinates.layers}.'
    )

  dx = jax_numpy_utils.diff(x, axis=axis)
  dx_axes = range(dx.ndim)
  inv_d𝜎 = 1 / coordinates.center_to_center
  inv_d𝜎_axes = [dx_axes[axis]]
  return einsum(dx, dx_axes, inv_d𝜎, inv_d𝜎_axes, dx_axes, precision='float32')  # pytype: disable=bad-return-type


@jax.named_call
def cumulative_sigma_integral(
    x: Array,
    coordinates: SigmaCoordinates,
    axis: int = -3,
    downward: bool = True,
    cumsum_method: str = 'dot',
    sharding: jax.sharding.NamedSharding | None = None,
) -> jax.Array:
  """Approximates the integral of a quantity `x` with respect to 𝜎.

  Uses a midpoint rule (https://en.wikipedia.org/wiki/Riemann_sum#Midpoint_rule)
  to approximate the integral

    ∫x d𝜎   𝜎= 0...b

  where 𝜎= 0...b if `downward` is True, and 𝜎= 1...b otherwise, and `b` ranges
  over `coordinates.boundaries`. Note that this integral is approximated
  differently to `log_sigma_integral`. This reflects the approach described in

    Durran, Dale. "Numerical Methods for Fluid Dynamics", 2010. § 8.6.3

  Args:
    x: an array of values with `x.shape[axis] == coordinates.layers`. These
      values are "located" at `coordinates` layer centers.
    coordinates: a `SigmaCoordinates` object describing the vertical coordinates
      to differentiate against. Must satisfy `coordinates.layers ==
      x.shape[axis]`.
    axis: the axis along which to approximate the derivative. Defaults to -3
      since this is the axis we typically use to index layers.
    downward: the direction of the integral. If True, the integral is taken
      "down" from the top of the atmosphere (𝜎 = 1). If False, it is taken "up"
      from the ground (𝜎 = 1).
    cumsum_method: method to use for calculating cumsum.
    sharding: optional parallel sharding for x.

  Returns:
    An array with the same shape as `x` containing approximate values of the
    integral of `x` from the 𝜎 = 0 to each layer's {lower, upper} boundary when
    `downward` is set to {True, False}.
  """
  if coordinates.layers != x.shape[axis]:
    raise ValueError(
        '`x.shape[axis]` must be equal to `coordinates.layers`;'
        f'got {x.shape[axis]} and {coordinates.layers}.'
    )
  x_axes = range(x.ndim)
  d𝜎 = coordinates.layer_thickness
  d𝜎_axes = [x_axes[axis]]
  xd𝜎 = einsum(x, x_axes, d𝜎, d𝜎_axes, x_axes)
  if downward:
    return jax_numpy_utils.cumsum(
        xd𝜎, axis, method=cumsum_method, sharding=sharding
    )
  else:
    return jax_numpy_utils.reverse_cumsum(
        xd𝜎, axis, method=cumsum_method, sharding=sharding
    )


@jax.named_call
def sigma_integral(
    x: Array,
    coordinates: SigmaCoordinates,
    axis: int = -3,
    keepdims: bool = True,
) -> jax.Array:
  """Calculate a full integral of a quantity `x` with respect to 𝜎."""
  if coordinates.layers != x.shape[axis]:
    raise ValueError(
        '`x.shape[axis]` must be equal to `coordinates.layers`;'
        f'got {x.shape[axis]} and {coordinates.layers}.'
    )
  x_axes = range(x.ndim)
  d𝜎 = coordinates.layer_thickness
  d𝜎_axes = [x_axes[axis]]
  xd𝜎 = einsum(x, x_axes, d𝜎, d𝜎_axes, x_axes)
  return xd𝜎.sum(axis=axis, keepdims=keepdims)


@jax.named_call
def cumulative_log_sigma_integral(
    x: Array,
    coordinates: SigmaCoordinates,
    axis: int = -3,
    downward: bool = True,
    cumsum_method: str = 'dot',
) -> jax.Array:
  """Approximates the integral of a quantity `x` with respect to log(𝜎).

  Uses the trapezoid rule (https://en.wikipedia.org/wiki/Trapezoidal_rule) to
  approximate the integral

    ∫x d(log(𝜎))

  where 𝜎= 0...c if `downward` is True, and 𝜎= 1...c otherwise, where `c` ranges
  over `coordinates.centers`. Note that this integral is approximated
  differently to `sigma_integral`. This reflects the approach described in

    Durran, Dale. "Numerical Methods for Fluid Dynamics", 2010. § 8.6.3

  Args:
    x: an array of values with `x.shape[axis] == coordinates.layers`. These
      values are "located" at `coordinates` layer centers.
    coordinates: a `SigmaCoordinates` object describing the vertical coordinates
      to differentiate against. Must satisfy `coordinates.layers ==
      x.shape[axis]`.
    axis: the axis along which to approximate the derivative. Defaults to -3
      since this is the axis we typically use to index layers.
    downward: the direction of the integral. If True, the integral is taken
      "down" from the top of the atmosphere (𝜎 = 1). If False, it is taken "up"
      from the ground (𝜎 = 1).
    cumsum_method: method to use for calculating cumsum.

  Returns:
    An array with the same shape as `x` containing approximate values of the
    integral of `x` from the "surface" (𝜎 = 1) up to each layer center.
  """
  if coordinates.layers != x.shape[axis]:
    raise ValueError(
        '`x.shape[axis]` must be equal to `coordinates.layers`;'
        f'got {x.shape[axis]} and {coordinates.layers}.'
    )
  # To integrate using the trapezoid rule, we linearly interpolate values of
  # `x`. The exception is between the surface (𝜎 = 1) and the center of the
  # first layer. In this interval, we assume a constant value of `x[-1]`.
  x_last = lax.slice_in_dim(x, -1, None, axis=axis)
  x_interpolated = (
      lax.slice_in_dim(x, 1, None, axis=axis)
      + lax.slice_in_dim(x, 0, -1, axis=axis)
  ) / 2
  integrand = jnp.concatenate([x_interpolated, x_last], axis=axis)
  integrand_axes = range(integrand.ndim)
  log𝜎 = jnp.log(coordinates.centers)
  dlog𝜎 = jnp.diff(log𝜎, append=0)
  dlog𝜎_axes = [integrand_axes[axis]]
  xd𝜎 = einsum(integrand, integrand_axes, dlog𝜎, dlog𝜎_axes, integrand_axes)
  if downward:
    return jax_numpy_utils.cumsum(xd𝜎, axis, method=cumsum_method)
  else:
    return jax_numpy_utils.reverse_cumsum(xd𝜎, axis, method=cumsum_method)


@jax.named_call
def centered_vertical_advection(
    w: Array,
    x: Array,
    coordinates: SigmaCoordinates,
    axis: int = -3,
    w_boundary_values: tuple[Array, Array] | None = None,
    dx_dsigma_boundary_values: tuple[Array, Array] | None = None,
) -> jnp.ndarray:
  """Compute vertical advection using 2nd order finite differences.

  Computes `-(w * ∂x/∂𝜎)[n]` at `coordinates.centers` with averaging.

  This procedure computes the expression using the averaging approximation:

  -(w * ∂x/∂𝜎)[n] ≈ -0.5 * (w[n+0.5] * ∂x/∂𝜎[n+0.5] + w[n-0.5] * ∂x/∂𝜎[n-0.5])

  To obtain values at all levels `w` and `∂x/∂𝜎` is padded with boundary values.
  By default boundary values for both terms are set to zero. Inputs are expected
  to be provided.

  Args:
    w: array of vertical velocities at `coordinates.internal_boundaries`.
    x: array of values at `coordinates.centers` that will be differentiated.
      Must be shape-compatible with `w` along axes other than `axis`.
    coordinates: object describing the vertical levels.
    axis: the axis corresponding to vertical (𝜎) values in `x` and `w`.
    w_boundary_values: optional top, bottom boundary values for `w` with shapes
      broacastable to `w` slices along `axis`.
    dx_dsigma_boundary_values: optional top, bottom boundary values for `∂x/∂𝜎`
      with shapes broacastable to `x` slices along `axis`.

  Returns:
    Values of `-(w * ∂x/∂𝜎)[n]` at level centers.
  """
  if w_boundary_values is None:
    w_slc_shape = _slice_shape_along_axis(w, axis)
    w_boundary_values = (
        jnp.zeros(w_slc_shape, dtype=jax.dtypes.canonicalize_dtype(w.dtype)),
        jnp.zeros(w_slc_shape, dtype=jax.dtypes.canonicalize_dtype(w.dtype)),
    )
  if dx_dsigma_boundary_values is None:
    x_slc_shape = _slice_shape_along_axis(x, axis)
    dx_dsigma_boundary_values = (
        jnp.zeros(x_slc_shape, dtype=jax.dtypes.canonicalize_dtype(x.dtype)),
        jnp.zeros(x_slc_shape, dtype=jax.dtypes.canonicalize_dtype(x.dtype)),
    )

  w_boundary_top, w_boundary_bot = w_boundary_values
  w = jnp.concatenate([w_boundary_top, w, w_boundary_bot], axis=axis)

  x_diff = centered_difference(x, coordinates, axis)
  x_diff_boundary_top, x_diff_boundary_bot = dx_dsigma_boundary_values
  x_diff = jnp.concatenate(
      [x_diff_boundary_top, x_diff, x_diff_boundary_bot], axis=axis
  )

  w_times_x_diff = w * x_diff
  return -0.5 * (
      lax.slice_in_dim(w_times_x_diff, 1, None, axis=axis)
      + lax.slice_in_dim(w_times_x_diff, 0, -1, axis=axis)
  )


@jax.named_call
def upwind_vertical_advection(
    w: Array,
    x: Array,
    coordinates: SigmaCoordinates,
    axis: int = -3,
) -> jnp.ndarray:
  """Compute vertical advection using 1st order upwinding."""
  w_slc_shape = _slice_shape_along_axis(w, axis)
  w_boundary_values = (
      jnp.zeros(w_slc_shape, dtype=jax.dtypes.canonicalize_dtype(w.dtype)),
      jnp.zeros(w_slc_shape, dtype=jax.dtypes.canonicalize_dtype(w.dtype)),
  )

  x_slc_shape = _slice_shape_along_axis(x, axis)
  dx_dsigma_boundary_values = (
      jnp.zeros(x_slc_shape, dtype=jax.dtypes.canonicalize_dtype(x.dtype)),
      jnp.zeros(x_slc_shape, dtype=jax.dtypes.canonicalize_dtype(x.dtype)),
  )

  # https://en.wikipedia.org/wiki/Upwind_scheme#Compact_form
  x_diff = centered_difference(x, coordinates, axis)

  w_boundary_top, w_boundary_bot = w_boundary_values
  w_up = jnp.concatenate([w_boundary_top, w], axis=axis)
  w_down = jnp.concatenate([w, w_boundary_bot], axis=axis)

  x_diff_boundary_top, x_diff_boundary_bot = dx_dsigma_boundary_values
  x_diff_up = jnp.concatenate([x_diff_boundary_top, x_diff], axis=axis)
  x_diff_down = jnp.concatenate([x_diff, x_diff_boundary_bot], axis=axis)
  # tendency (i.e. r.h.s. has a negative sign).
  return -(jnp.maximum(w_up, 0) * x_diff_up +
           jnp.minimum(w_down, 0) * x_diff_down)


# pylint: enable=invalid-name



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Code for generating conditions for shallow water system.

Here we provide steady geostrophic flow and initial state from barotropic
instability test case [1].

  [1]: Galewsky, Joseph, Richard K. Scott, and Lorenzo M. Polvani.
       "An initial-value problem for testing numerical models of the global
       shallow-water equations."
       Tellus A: Dynamic Meteorology and Oceanography 56, no. 5 (2004): 429-440.
"""

import functools
from typing import Any, Callable, NamedTuple

from dinosaur import coordinate_systems
from dinosaur import scales
from dinosaur import shallow_water
from dinosaur import spherical_harmonic
from dinosaur import typing
from dinosaur import xarray_utils

import jax
import jax.numpy as jnp
import numpy as np


units = scales.units
Array = typing.Array
Numeric = typing.Numeric


def subtract_longitudes(a, b):
  """Computes the difference `a - b`, accounting for periodicity."""
  return jnp.mod(a - b + jnp.pi, 2 * jnp.pi) - jnp.pi


def one_layer(
    u: Array,
    grid: spherical_harmonic.Grid
) -> shallow_water.State:
  """Simple steady state for one-layer rotating SWE.

  Args:
    u: longitudinal velocities u(θ) in nodal representation.
    grid: object describing horizontal spatial discretization.

  Returns:
    A steady state solution to the shallow water equations. The steady state
    satisfies

      ∂(Φ + ½u²)/∂θ + (ζ + f) u = 0

    where ζ is vorticity and f is the Coriolis force.
  """
  # Construct a velocity field that has zonal (eastward) velocity specified
  # by `u` at each Gaussian quadrature point. The meridional (northward)
  # velocity is uniformly zero.
  sec_lat = 1 / grid.cos_lat
  lon, _ = grid.nodal_axes
  u = jnp.ones_like(lon)[..., jnp.newaxis] * u
  sec_lat_u = u * sec_lat
  vorticity = -grid.sec_lat_d_dlat_cos2(grid.to_modal(sec_lat_u))
  f = shallow_water.get_coriolis(grid)
  total_vorticity = grid.to_nodal(vorticity) + f
  potential_plus_energy = -grid.inverse_laplacian(
      grid.sec_lat_d_dlat_cos2(grid.to_modal(sec_lat_u * total_vorticity)))
  potential = potential_plus_energy - grid.to_modal(u**2 / 2)
  potential = potential.at[0, 0].set(0)
  return shallow_water.State(vorticity=vorticity,
                             divergence=jnp.zeros_like(vorticity),
                             potential=potential)


def multi_layer(
    u: Array,
    density: Array,
    coords: coordinate_systems.CoordinateSystem,
) -> shallow_water.State:
  """Simple steady state for n-layer rotating SWE.

  Args:
    u: longitudinal velocities u(θ) in nodal representation.
    density: a non-decreasing vector of densities, starting from the top layer.
    coords: horizontal and vertical descritization.

  Returns:
    The steady state solution to the multi-layer SWE, with vᵢ = 0, which
    satisfies

      ∂(RΦ + ½u²)/∂θ + (ζ + f) u = 0

    where `R[i, j] = min(⍴[j] / ⍴[i], 1)`, ζ is vorticity and f is the
    Coriolis force.
  """
  s = jax.vmap(functools.partial(one_layer, grid=coords.horizontal))(u)
  n_layers = coords.vertical.layers
  density_ratios = shallow_water.get_density_ratios(density) + np.eye(n_layers)
  potential = s.potential
  if n_layers > 1:
    # TODO(jamieas): consider just inverting `density_ratios`, as this should be
    # a small matrix.
    flat_potential = jnp.reshape(s.potential, (n_layers, -1))
    potential = jnp.reshape(jnp.linalg.solve(density_ratios, flat_potential),
                            s.potential.shape)
  return shallow_water.State(s.vorticity, s.divergence, potential)


class BarotropicInstabilityParameters(NamedTuple):
  """Parameters of initial state for a barotropic instability test case."""

  # Parameters describing the location and magnitude of the velocity 'jet'.
  jet_northern_lat: Numeric
  jet_southern_lat: Numeric
  jet_max_velocity: Numeric

  # The mean thickness of the atmosphere.
  mean_height: Numeric

  # Parameters describing the latitude, longitude, size and size of the 'bump'
  # that is added to the height field.
  bump_lon_location: Numeric
  bump_lon_scale: Numeric
  bump_lat_location: Numeric
  bump_lat_scale: Numeric
  bump_height_scale: Numeric


def get_default_parameters() -> BarotropicInstabilityParameters:
  """Returns the parameters from the original paper, cited above."""
  return BarotropicInstabilityParameters(
      jet_northern_lat=np.array(np.pi / 2 - np.pi / 7) * units.radian,
      jet_southern_lat=np.array(np.pi / 7) * units.radian,
      jet_max_velocity=np.array(80.) * units.m / units.s,
      mean_height=np.array(10_000.) * units.m,
      bump_lon_location=np.array(0.) * units.radian,
      bump_lon_scale=np.array(1 / 3) * units.radian,
      bump_lat_location=np.array(np.pi / 4) * units.radian,
      bump_lat_scale=np.array(1 / 15) * units.radian,
      bump_height_scale=np.array(120.) * units.m)


def get_random_parameters(
    key: jnp.ndarray,
    default_parameters: BarotropicInstabilityParameters,
) -> BarotropicInstabilityParameters:
  """Return random parameters for the barotropic instability test case."""
  # Currently, we only randomize the latitude of the 'bump', and it is required
  # to be located 'within' the jet.
  jet_width = (default_parameters.jet_northern_lat
               - default_parameters.jet_southern_lat)
  random_bump_lat = (default_parameters.jet_southern_lat
                     + jax.random.beta(key, 4, 4) * jet_width)
  return default_parameters._replace(bump_lat_location=random_bump_lat)


def get_zonal_velocity(
    latitude: Numeric,
    parameters: BarotropicInstabilityParameters,
) -> Array:
  """Returns the zonal velocity based on `parameters` for `latitude`."""
  inside_jet = ((latitude > parameters.jet_southern_lat)
                * (latitude < parameters.jet_northern_lat))
  velocity_shape = jnp.exp(1 / (latitude - parameters.jet_southern_lat)
                           / (latitude - parameters.jet_northern_lat))
  normalizer = jnp.exp(
      -4 / (parameters.jet_northern_lat - parameters.jet_southern_lat)**2)
  velocity_inside_jet = (
      parameters.jet_max_velocity / normalizer * velocity_shape)

  return jnp.where(inside_jet, velocity_inside_jet, 0)


def get_height(
    longitude: Numeric,
    latitude: Numeric,
    parameters: BarotropicInstabilityParameters
):
  """Returns the height profile of the bump specified in `parameters`."""
  bump_shape = jnp.cos(latitude) * jnp.exp(
      -(subtract_longitudes(parameters.bump_lon_location, longitude)
        / parameters.bump_lon_scale) ** 2
      -((parameters.bump_lat_location - latitude)
        / parameters.bump_lat_scale) ** 2)
  return parameters.bump_height_scale * bump_shape


def barotropic_instability_tc(
    coords: coordinate_systems.CoordinateSystem,
    physics_specs: Any,
) -> tuple[Callable[..., shallow_water.State], typing.AuxFeatures]:
  """Returns a function that generates random states and static features.

  This function implements initial states for barotropic instability test case.
  The returned features includes mean potential at the corresponding levels.

  Args:
    coords: horizontal and vertical descritization.
    physics_specs: physics specs describing global physical parameters.

  Returns:
    random_state_fn: function that generates a randomized initial state.
    aux_features: static auxiliary data to be provided to dynamical models.
  """
  _, sin_lat = coords.horizontal.nodal_mesh
  lat = np.arcsin(sin_lat)
  default_parameters = get_default_parameters()

  def random_state_fn(rng_key: jnp.ndarray) -> shallow_water.State:
    parameters = get_random_parameters(rng_key, default_parameters)
    # make sure that all parameters are non-dimensionalized.
    parameters = jax.tree.map(physics_specs.nondimensionalize, parameters)
    # The initial condition is computed by findind a steady state solution and
    # then adding a small 'bump' to the potential.
    zonal_velocity = jnp.stack([get_zonal_velocity(lat, parameters)
                                for _ in range(coords.vertical.layers)])
    steady = multi_layer(zonal_velocity, physics_specs.densities, coords)
    bump_potential = coords.horizontal.to_modal(
        get_height(*coords.horizontal.nodal_mesh, parameters) * physics_specs.g)
    initial_state = shallow_water.State(
        vorticity=steady.vorticity,
        potential=(steady.potential + bump_potential),
        divergence=steady.divergence)
    return initial_state

  top_potential = physics_specs.nondimensionalize(
      default_parameters.mean_height) * physics_specs.g
  reference_potential = np.linspace(
      0, top_potential, 1 + coords.vertical.layers, endpoint=True)[1:]
  aux_features = {xarray_utils.REF_POTENTIAL_KEY: reference_potential}
  return random_state_fn, aux_features



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Code for describing units and non-dimensionalizing quantities."""

from collections import abc
from typing import Iterator, Protocol

import jax.numpy as jnp
import numpy as np
import pint

units = pint.UnitRegistry(autoconvert_offset_to_baseunit=True)

Quantity = units.Quantity
Unit = units.Unit
UnitsContainer = pint.util.UnitsContainer

Array = np.ndarray | jnp.ndarray
Numeric = Array | float | int

#
# Physical constants.
#

# The radius of the earth.
RADIUS = 6.37122e6 * units.m

# The rotation rate if the Earth in radians per second, often denoted Ω.
ANGULAR_VELOCITY = OMEGA = 7.292e-5 / units.s

# Acceleration due to gravity on Earth.
GRAVITY_ACCELERATION = 9.80616 * units.m / units.s ** 2

# Specific heat capacity at constant pressure.
ISOBARIC_HEAT_CAPACITY = 1004 * units.J / units.kilogram / units.degK

# Specific heat capacity of water vapor at constant pressure.
# value taken for T=275 from:
# https://www.engineeringtoolbox.com/water-vapor-d_979.html
WATER_VAPOR_CP = 1859 * units.J / units.kilogram / units.degK

# The mass of the dry atmosphere.
MASS_OF_DRY_ATMOSPHERE = 5.18e18 * units.kg

# The ratio of the ideal gas constant to the isobaric specific heat capacity of
# a diatomic ideal gas, often denoted κ. This value corresponds to a heat
# capacity ratio ɣ = 7 / 5. Note that this quantity is dimensionless, and this
# "unit" is included for consistency with other constants.
KAPPA = 2 / 7 * units.dimensionless

# The Latent Heat of Vaporization for water assuming T = 273.15 K
# Used to calculate enthalpy and MSE budgets as well as other moisture values
# https://glossary.ametsoc.org/wiki/Latent_heat
LATENT_HEAT_OF_VAPORIZATION = 2.501e6 * units.J / units.kilogram

# Ideas gas constant for dry air (di-atomic). For `KAPPA == 1004` this value is
# approximately 287.07.
IDEAL_GAS_CONSTANT = ISOBARIC_HEAT_CAPACITY * KAPPA

# Ideal gas constant for air with water vapor included
# Generally approximated to 461.
IDEAL_GAS_CONSTANT_H20 = 461. * units.J / units.kilogram / units.degK

# Freezing point of Temperature in Kelvin
# Used to convert between Kelvin and degrees Celsius
T_FREEZING = 273.15 * units.degK

# Pressure of one atmosphere (standard pressure)
# used to calculate potential temperature away from surface
REFERENCE_PRESSURE = 101325. *units.pascal

# Density of water in SI units
WATER_DENSITY = 997 * units.kg / units.m ** 3

#
# Code for defining scales and non-dimensionalizing quantities.
#


def parse_units(units_str: str) -> Quantity:
  if units_str in {'(0 - 1)', '%', '~'}:
    units_str = 'dimensionless'
  return units.parse_expression(units_str)


def _get_dimension(quantity: Quantity) -> str:
  """Asserts `quantity` has a single dimension and returns that dimension."""
  exponents = list(quantity.dimensionality.values())
  if len(quantity.dimensionality) != 1 or exponents[0] != 1:
    raise ValueError('All scales must describe a single dimension;'
                     f'got dimensionality {quantity.dimensionality}')
  return str(quantity.dimensionality)


class ScaleProtocol(Protocol):
  """A protocol class for `Scale` objects that perform nondimensionalization."""

  def nondimensionalize(self, quantity: Quantity) -> Numeric:
    """Converts a `pint.Quantity` to a non-dimensional value."""
    ...

  def dimensionalize(self, value: Numeric, unit: Unit) -> Quantity:
    """Converts non-dimensional `value` to a `pint.Quantity` with `unit`."""
    ...


class Scale(abc.Mapping):
  """A `Scale` converts values to and from dimensionless quantities."""

  def __init__(self, *scales: Quantity):
    """Initializes a `Scale` from the given `Quantity`s.

    Example usage:

    ```python
    # The radius of the Earth, in meters.
    RADIUS = 6.37122e6 * units.m

    # The rotation rate of the earth, in radians per second.
    OMEGA = 7.292e-5 / units.sec

    scale = Scale(RADIUS, 1 / 2 / OMEGA)

    g = 9.8 units.m / units.s ** 2
    dimensionless_g = scale.nondimensionalize(g)
    ```

    Args:
      *scales: `pint.Quantity`s indicating scales. The dimensionality of
        the argument must be one of [current], [length], [luminosity], [mass],
        [printing_unit], [substance], [temperature], [time]. In particular, each
        scale must not have a 'compound' unit such as 'm / s'.
    """
    self._scales = dict()
    for quantity in scales:
      dimension = _get_dimension(quantity)
      if dimension in self._scales:
        raise ValueError(f'Got duplicate scales for dimension {dimension}.')
      self._scales[_get_dimension(quantity)] = quantity.to_base_units()

  def __getitem__(self, key: str) -> Quantity:
    return self._scales[key]

  def __iter__(self) -> Iterator[str]:
    return iter(self._scales)

  def __len__(self) -> int:
    return len(self._scales)

  def __repr__(self) -> str:
    return '\n'.join(f'{dimension}: {quantity}'
                     for dimension, quantity in self._scales.items())

  def _scaling_factor(self,
                      dimensionality: pint.util.UnitsContainer) -> Quantity:
    """Returns the value used to scale quantities of given dimensionality."""
    factor = Quantity(1)
    for dimension, exponent in dimensionality.items():
      quantity = self._scales.get(dimension)
      if quantity is None:
        raise ValueError(f'No scale has been set for {dimension}.')
      factor *= quantity ** exponent
    assert factor.check(dimensionality)
    return factor

  def nondimensionalize(self, quantity: Quantity) -> Numeric:
    """Converts a `pint.Quantity` to a non-dimensional value.

    Args:
      quantity: a `pint.Quantity` to be converted to a dimensionless quantity.
        The items in `quantity.dimensionality` should be keys in this `Scale`.

    Returns:
      A dimensionless value corresponding to the rescaled `quantity`.
    """
    scaling_factor = self._scaling_factor(quantity.dimensionality)
    nondimensionalized = (quantity / scaling_factor).to(units.dimensionless)
    return nondimensionalized.magnitude

  def dimensionalize(self, value: Numeric, unit: Unit) -> Quantity:
    """Scales `value` by the appropriate factor and attached `unit`.

    Args:
      value: a dimensionless value representing a quantity corresponding to
        `unit`.
      unit: the desired units for the returned `Quantity`.

    Returns:
      A `Quantity` corresponding to `value` after rescaling.
    """
    scaling_factor = self._scaling_factor(unit.dimensionality)
    dimensionalized = value * scaling_factor
    return dimensionalized.to(unit)  # pytype: disable=attribute-error  # jax-ndarray


DEFAULT_SCALE = Scale(RADIUS,              # length
                      1 / 2 / OMEGA,       # time
                      1 * units.kilogram,  # mass
                      1 * units.degK)      # temperature

ATMOSPHERIC_SCALE = Scale(
    RADIUS,                  # length
    1 / 2 / OMEGA,           # time
    MASS_OF_DRY_ATMOSPHERE,  # mass
    1 * units.degK)          # temperature



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Integration tests for primitive_equations."""
import functools

from absl.testing import absltest
import chex
from dinosaur import coordinate_systems
from dinosaur import primitive_equations
from dinosaur import primitive_equations_states
from dinosaur import pytree_utils
from dinosaur import scales
from dinosaur import sigma_coordinates
from dinosaur import spherical_harmonic
from dinosaur import time_integration
from dinosaur import xarray_utils
import jax
from jax import config  # pylint: disable=g-importing-member
import jax.numpy as jnp
import numpy as np


def make_coords(
    max_wavenumber: int,
    num_layers: int,
    spmd_mesh: jax.sharding.Mesh | None,
    spherical_harmonics_impl,
) -> coordinate_systems.CoordinateSystem:
  return coordinate_systems.CoordinateSystem(
      spherical_harmonic.Grid.with_wavenumbers(
          longitude_wavenumbers=max_wavenumber + 1,
          spherical_harmonics_impl=spherical_harmonics_impl,
      ),
      sigma_coordinates.SigmaCoordinates.equidistant(num_layers),
      spmd_mesh=spmd_mesh,
  )


def make_initial_state(coords, physics_specs):
  initial_state_fn, _ = primitive_equations_states.steady_state_jw(
      coords, physics_specs
  )
  state = initial_state_fn()
  state = state + primitive_equations_states.baroclinic_perturbation_jw(
      coords, physics_specs
  )
  state.tracers = {
      'specific_humidity': primitive_equations_states.gaussian_scalar(
          coords, physics_specs
      )
  }
  state.sim_time = 0.0
  return state


def make_dycore_sim_fn(coords, physics_specs, num_hours):
  _, aux_features = primitive_equations_states.steady_state_jw(
      coords, physics_specs
  )
  ref_temps = aux_features[xarray_utils.REF_TEMP_KEY]
  orography = np.zeros(coords.horizontal.modal_shape, np.float32)
  eq = primitive_equations.MoistPrimitiveEquations(
      ref_temps, orography, coords, physics_specs
  )
  dt = physics_specs.nondimensionalize(30 * scales.units.minute)
  step_fn = time_integration.imex_rk_sil3(eq, time_step=dt)
  sim_fn = time_integration.repeated(step_fn, 2 * num_hours)
  return jax.jit(sim_fn)


def assert_allclose(actual, desired, *, atol=0, range_tol=0, **kwargs):
  atol = max(atol, range_tol * (desired.max() - desired.min()))
  return np.testing.assert_allclose(actual, desired, atol=atol, **kwargs)


def assert_states_close(state0, state1, *, atol=0, range_tol=0, **kwargs):
  for field in state0.fields:
    if field.name == 'tracers':
      for tracer_name in state0.tracers.keys():
        assert_allclose(
            state0.tracers[tracer_name],
            state1.tracers[tracer_name],
            atol=atol,
            range_tol=range_tol,
            err_msg=f'Mismatch in tracer {tracer_name}:',
            **kwargs,
        )
    else:
      assert_allclose(
          getattr(state0, field.name),
          getattr(state1, field.name),
          atol=atol,
          range_tol=range_tol,
          err_msg=f'Mismatch in {field}:',
          **kwargs,
      )


def pad_state(state, distributed_coords):
  def f(x):
    pad_x, pad_y = (
        distributed_coords.horizontal.spherical_harmonics.modal_padding
    )
    return jnp.pad(x, [(0, 0), (0, pad_x), (0, pad_y)])

  return pytree_utils.tree_map_over_nonscalars(f, state)


def trim_state(state, distributed_coords):
  def f(x):
    pad_x, pad_y = (
        distributed_coords.horizontal.spherical_harmonics.modal_padding
    )
    return x[..., : -pad_x or None, : -pad_y or None]

  return pytree_utils.tree_map_over_nonscalars(f, state)


class IntegrationTest(absltest.TestCase):

  def test_distributed_simulation_consistency(self):
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    coords = make_coords(
        max_wavenumber=31,
        num_layers=8,
        spmd_mesh=None,
        spherical_harmonics_impl=spherical_harmonic.FastSphericalHarmonics,
    )
    init_state = make_initial_state(coords, physics_specs)
    sim_fn = make_dycore_sim_fn(coords, physics_specs, num_hours=1)
    non_distributed_state = sim_fn(init_state)
    non_distributed_nodal = coords.horizontal.to_nodal(non_distributed_state)

    with self.subTest('vertical sharding'):
      devices = np.array(jax.devices()[:2]).reshape((2, 1, 1))
      mesh = jax.sharding.Mesh(devices, axis_names=['z', 'x', 'y'])
      distributed_coords = make_coords(
          max_wavenumber=31,
          num_layers=8,
          spmd_mesh=mesh,
          spherical_harmonics_impl=spherical_harmonic.FastSphericalHarmonics,
      )
      distributed_init_state = pad_state(init_state, distributed_coords)
      distributed_sim_fn = make_dycore_sim_fn(
          distributed_coords, physics_specs, num_hours=1
      )
      distributed_state = distributed_sim_fn(distributed_init_state)
      distributed_state = trim_state(distributed_state, distributed_coords)
      distributed_nodal = coords.horizontal.to_nodal(distributed_state)

      assert_states_close(
          non_distributed_nodal, distributed_nodal, rtol=1e-6, range_tol=1e-6
      )

    with self.subTest('horizontal sharding'):
      devices = np.array(jax.devices()[:4]).reshape((1, 2, 2))
      mesh = jax.sharding.Mesh(devices, axis_names=['z', 'x', 'y'])
      distributed_coords = make_coords(
          max_wavenumber=31,
          num_layers=8,
          spmd_mesh=mesh,
          spherical_harmonics_impl=spherical_harmonic.FastSphericalHarmonics,
      )
      distributed_init_state = pad_state(init_state, distributed_coords)
      distributed_sim_fn = make_dycore_sim_fn(
          distributed_coords, physics_specs, num_hours=1
      )
      distributed_state = distributed_sim_fn(distributed_init_state)
      distributed_state = trim_state(distributed_state, distributed_coords)
      distributed_nodal = coords.horizontal.to_nodal(distributed_state)

      assert_states_close(
          non_distributed_nodal, distributed_nodal, rtol=1e-6, range_tol=1e-6
      )

  def test_real_vs_fast_spherical_harmonics(self):
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()

    real_coords = make_coords(
        max_wavenumber=31,
        num_layers=8,
        spmd_mesh=None,
        spherical_harmonics_impl=spherical_harmonic.RealSphericalHarmonics,
    )
    fast_coords = make_coords(
        max_wavenumber=31,
        num_layers=8,
        spmd_mesh=None,
        spherical_harmonics_impl=functools.partial(
            spherical_harmonic.FastSphericalHarmonics,
            transform_precision='float32',
        ),
    )

    real_init_state = make_initial_state(real_coords, physics_specs)
    fast_init_state = make_initial_state(fast_coords, physics_specs)

    real_init_nodal = real_coords.horizontal.to_nodal(real_init_state)
    fast_init_nodal = fast_coords.horizontal.to_nodal(fast_init_state)

    with self.subTest('initial conditions'):
      assert_states_close(
          real_init_nodal, fast_init_nodal, rtol=1e-6, range_tol=1e-6
      )

    sim_fn = make_dycore_sim_fn(real_coords, physics_specs, num_hours=1)
    real_out_state = sim_fn(real_init_state)
    real_out_nodal = real_coords.horizontal.to_nodal(real_out_state)

    sim_fn = make_dycore_sim_fn(fast_coords, physics_specs, num_hours=1)
    fast_out_state = sim_fn(fast_init_state)
    fast_out_nodal = fast_coords.horizontal.to_nodal(fast_out_state)

    with self.subTest('evolved state'):
      assert_states_close(
          real_out_nodal, fast_out_nodal, rtol=1e-4, range_tol=1e-5
      )


if __name__ == '__main__':
  chex.set_n_cpu_devices(8)
  config.update('jax_traceback_filtering', 'off')
  absltest.main()



# Copyright 2023 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Multi-layer shallow water equations."""

from __future__ import annotations

import dataclasses
import functools
from typing import Callable, Sequence, Type

from dinosaur import coordinate_systems
from dinosaur import scales
from dinosaur import spherical_harmonic
from dinosaur import time_integration
from dinosaur import typing

import jax
import jax.numpy as jnp
import numpy as np
import tree_math

units = scales.units

Array = typing.Array
Numeric = typing.Numeric
Quantity = typing.Quantity

FilterFn = typing.FilterFn
InverseFn = typing.InverseFn
StateFn = typing.StateFn
StepFn = typing.StepFn

SCALE = scales.DEFAULT_SCALE


# All `einsum`s should be done at highest available precision.
einsum = functools.partial(jnp.einsum, precision=jax.lax.Precision.HIGHEST)

#  =============================================================================
#  Data Structures
#
#  Data classes that describe the state, scale and parameters of the system.
#  =============================================================================


@tree_math.struct
class State:
  """Records the state of a system described by the shallow water equations."""
  vorticity: Array
  divergence: Array
  potential: Array


@dataclasses.dataclass(frozen=True)
class ShallowWaterSpecs:
  """Records scales and physical constants used in the shallow water equations.

  By default uses units in which the radius and angular_velocity are set to `1`.

  Attributes:
    densities: a non-decreasing vector of densities, starting from the top.
    radius: the non-dimensionalized radius of the domain.
    angular_velocity: the non-dimensionalized angular velocity of the rotating
      domain.
    gravity_acceleration: the non-dimensionalized value of gravitational
      acceleration.
    scale: an instance implementing `ScaleProtocol` that will be used to
      (non-)dimensionalize quantities.
  """
  densities: Array
  radius: float
  angular_velocity: float
  gravity_acceleration: float
  scale: scales.ScaleProtocol

  @property
  def g(self) -> float:
    """Alias for `gravity_acceleration`."""
    return self.gravity_acceleration

  @property
  def num_layers(self) -> int:
    """Returns number of layers in the equation."""
    return self.densities.size

  def nondimensionalize(self, quantity: Quantity) -> Numeric:
    """Non-dimensionalizes and rescales `quantity`."""
    return self.scale.nondimensionalize(quantity)

  def dimensionalize(self, value: Numeric, unit: units.Unit) -> Quantity:
    """Rescales and adds units to the given non-dimensional value."""
    return self.scale.dimensionalize(value, unit)

  @classmethod
  def from_si(
      cls: Type[ShallowWaterSpecs],
      densities: Quantity = np.ones(1) * scales.WATER_DENSITY,
      radius_si: Quantity = scales.RADIUS,
      angular_velocity_si: Quantity = scales.ANGULAR_VELOCITY,
      gravity_acceleration_si: Quantity = scales.GRAVITY_ACCELERATION,
      scale: scales.ScaleProtocol = scales.DEFAULT_SCALE) -> ShallowWaterSpecs:
    """Constructs `ShallowWaterSpecs` from SI constants."""
    return cls(scale.nondimensionalize(densities),
               scale.nondimensionalize(radius_si),
               scale.nondimensionalize(angular_velocity_si),
               scale.nondimensionalize(gravity_acceleration_si),
               scale)


#  =============================================================================
#  Helper Functions
#
#  Functions used to compute individual terms and intermediate values for the
#  primitive equations.
#  =============================================================================


def state_to_nodal(state: State, grid: spherical_harmonic.Grid) -> State:
  """Converts a state to the spatial/nodal basis."""
  return jax.tree.map(
      lambda x: grid.to_nodal(grid.clip_wavenumbers(x)), state)


def state_to_modal(state: State, grid: spherical_harmonic.Grid) -> State:
  """Converts a state to the spectral/modal basis."""
  return jax.tree.map(grid.to_modal, state)


def get_density_ratios(density: Array) -> np.ndarray:
  """Computes density ratios used to compute interactions between layers.

  Args:
    density: a vector of layer densities, beginning from the top. These values
      must be non-decreasing.

  Returns:
    An array `D` such that

                 density[i] / density[j]  if i < j
      D[i, j] =  0                        if i = j
                 1                        if i > j
  """
  ratios = np.minimum(density / density[..., np.newaxis], 1)
  np.fill_diagonal(ratios, 0)
  return ratios


def get_coriolis(grid: spherical_harmonic.Grid) -> np.ndarray:
  """Returns an array of coriolis forces in the spatial basis."""
  _, sin_lat = grid.nodal_mesh
  return sin_lat


#  =============================================================================
#  The `ShallowWaterEquations` Class
#
#  The `ShallowWaterEquations` class expresses the shallow water equations in a
#  form that is appropriate for semi-implicit time stepping.
#  =============================================================================


@dataclasses.dataclass
class ShallowWaterEquations(time_integration.ImplicitExplicitODE):
  """A semi-implicit description of the shallow water equations.

  See go/shallow-water for more details.

  Attributes:
    coords: horizontal and vertical descritization.
    physics_specs: an object describing the scales and physical constants.
    orography: an array of shape [latitudinal_wavenumbers, total_wavenumbers]
      describing the topography.
    reference_potential: an array of shape [layers] holding mean geopotential.
  """
  coords: coordinate_systems.CoordinateSystem
  physics_specs: ShallowWaterSpecs
  orography: Array
  reference_potential: Array

  @property
  def coriolis_parameter(self) -> Array:
    """Returns the value `2Ω sin(θ)` associated with Coriolis force."""
    _, sin_lat = self.coords.horizontal.nodal_mesh
    return 2 * self.physics_specs.angular_velocity * sin_lat

  @property
  def density_ratios(self) -> Array:
    """Returns `density_ratios` with spatial dimensions appended."""
    return get_density_ratios(self.physics_specs.densities)

  @property
  def ref_potential(self) -> Array:
    """Returns `reference_potential` with spatial dimensions appended."""
    return self.reference_potential[..., np.newaxis, np.newaxis]

  def explicit_terms(self, state: State) -> State:
    """Computes explicit tendencies of the shallow water equations."""
    # we stack two components of the velocity to transform them together.
    u = jnp.stack(spherical_harmonic.get_cos_lat_vector(
        state.vorticity, state.divergence, self.coords.horizontal))

    # Switch to physical coordinates for spatial point-wise operations
    nodal_u = self.coords.horizontal.to_nodal(u)
    nodal_state = state_to_nodal(state, self.coords.horizontal)

    total_vorticity = nodal_state.vorticity + self.coriolis_parameter

    sec2_lat = self.coords.horizontal.sec2_lat
    nodal_b = nodal_u * total_vorticity * sec2_lat
    nodal_g = nodal_u * nodal_state.potential * sec2_lat
    nodal_e = (nodal_u * nodal_u).sum(0) * sec2_lat / 2

    # Stack and unstack values to perform a single transform
    bge_nodal = jnp.concatenate(
        [nodal_b, nodal_g, jnp.expand_dims(nodal_e, axis=0)], axis=0)
    bge = self.coords.horizontal.to_modal(bge_nodal)
    b, g, e = jnp.split(bge, [2, 4], axis=0)
    e = jnp.squeeze(e, axis=0)

    # Pressure gradients are computed as weighted sums across layers.
    # Note that this is the only interaction between layers.
    p = einsum('ab,...bml->...aml',
               self.density_ratios,
               state.potential)
    if self.orography is not None:
      p = p + self.orography

    explicit_vorticity = self.coords.horizontal.clip_wavenumbers(
        -self.coords.horizontal.div_cos_lat(b)
    )
    explicit_divergence = self.coords.horizontal.clip_wavenumbers(
        -self.coords.horizontal.laplacian(p + e) +
        self.coords.horizontal.curl_cos_lat(b)
    )
    explicit_potential = self.coords.horizontal.clip_wavenumbers(
        -self.coords.horizontal.div_cos_lat(g)
    )
    return State(explicit_vorticity, explicit_divergence, explicit_potential)

  def implicit_terms(self, state: State) -> State:
    """Returns the implicit terms of the shallow water equations."""
    return State(
        vorticity=jnp.zeros_like(state.vorticity),
        divergence=-self.coords.horizontal.laplacian(state.potential),
        potential=-self.ref_potential * state.divergence)

  def implicit_inverse(self, state: State, step_size: float) -> State:
    """Computes the inverse `(1 - step_size * implicit_terms)⁻¹."""
    inverse_schur_complement = 1 / (
        1 - step_size ** 2 * self.ref_potential *
        self.coords.horizontal.laplacian_eigenvalues)
    return State(
        vorticity=state.vorticity,
        divergence=inverse_schur_complement * (
            state.divergence -
            step_size * self.coords.horizontal.laplacian(state.potential)
        ),
        potential=inverse_schur_complement * (
            -step_size * self.ref_potential * state.divergence + state.potential
        )
    )


def shallow_water_leapfrog_step(
    coords: coordinate_systems.CoordinateSystem,
    dt: float,
    physics_specs: ShallowWaterSpecs,
    mean_potential: np.ndarray,
    orography: Array | None = None,
    alpha: float = 0.5
) -> typing.TimeStepFn:
  """Returns a step function based on semi-implicit leapfrog integrator.

  Args:
    coords: horizontal and vertical descritization.
    dt: the size of the timestep used for integration.
    physics_specs: an `PrimitiveEquationSpecs` object describing the scales and
      physical constants used in the primitive equations.
    mean_potential: a vector of mean geopotentials g · h for each layer,
      starting from the top.
    orography: the geopotential g · h corresponding to the orography
      underlying the simulation. Must be in the spectral/modal basis.
    alpha: a parameter used to weight previous and future terms in the implicit
      portion of the equation: `f_i(alpha * future + (1 - alpha) * previous)`
  Returns:
    A function that computes a single time step of the shallow water equations.
    The returned function takes `state_0` and `state_1` states and returns the
    next state.
  """
  shallow_water_ode = ShallowWaterEquations(
      coords, physics_specs, orography, mean_potential)
  return time_integration.semi_implicit_leapfrog(shallow_water_ode, dt, alpha)


def shallow_water_leapfrog_trajectory(
    coords: coordinate_systems.CoordinateSystem,
    dt: float,
    physics_specs: ShallowWaterSpecs,
    inner_steps: int,
    outer_steps: int,
    mean_potential: np.ndarray,
    orography: Array | None = None,
    filters: Sequence[typing.PyTreeStepFilterFn] = (),
    alpha: float = 0.5,
) -> typing.TrajectoryFn:
  """Returns a trajectory function for shallow water equations."""
  step_fn = shallow_water_leapfrog_step(
      coords, dt, physics_specs, mean_potential, orography, alpha)
  step_fn = time_integration.step_with_filters(step_fn, filters)
  post_process_fn = lambda x: x[0]
  trajectory_fn = time_integration.trajectory_from_step(
      step_fn, outer_steps, inner_steps, post_process_fn=post_process_fn)
  return trajectory_fn


def default_filters(
    grid: spherical_harmonic.Grid,
    dt: float,
) -> Sequence[typing.PyTreeStepFilterFn]:
  """Returns standard filters for leapfrog integration of shallow water Eqs."""
  return (
      time_integration.exponential_leapfrog_step_filter(grid, dt),
      time_integration.robert_asselin_leapfrog_filter(0.05),
  )
